"""
This code is based on the work of James Chapman from cca-zoo.
Source: https://github.com/jameschapman19/cca_zoo

The original code is licensed under the MIT License.

Copyright (c) 2020 onward James Chapman
"""

from abc import abstractmethod
from datetime import datetime
from typing import Hashable, Sequence

import dask.array as da
import numpy as np
import xarray as xr
from scipy.linalg import eigh
from sklearn.base import BaseEstimator
from sklearn.utils.validation import FLOAT_DTYPES
from typing_extensions import Self

from .._version import __version__
from ..preprocessing.preprocessor import Preprocessor
from ..single import EOF
from ..utils.data_types import DataArray, DataList, DataObject
from ..utils.sanity_checks import assert_not_complex


def _check_parameter_number(parameter_name: str, parameter, n_views: int):
    if len(parameter) != n_views:
        raise ValueError(
            f"number of views passed should match number of parameter {parameter_name}"
            f"len(views)={n_views} and "
            f"len({parameter_name})={len(parameter)}"
        )


def _process_parameter(parameter_name: str, parameter, default, n_views: int):
    if parameter is None:
        parameter = [default] * n_views
    elif not isinstance(parameter, (list, tuple)):
        parameter = [parameter] * n_views
    _check_parameter_number(parameter_name, parameter, n_views)
    return parameter


class CCABaseModel(BaseEstimator):
    def __init__(
        self,
        n_modes: int = 10,
        use_coslat: bool = False,
        check_nans: bool = True,
        pca: bool = False,
        variance_fraction: float = 0.99,
        init_pca_modes: int | float = 0.75,
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: str = "feature",
    ):
        self.sample_name = sample_name
        self.feature_name = feature_name
        self.n_modes = n_modes
        self.use_coslat = use_coslat
        self.pca = pca
        self.compute = compute
        self.variance_fraction = variance_fraction
        self.init_pca_modes = init_pca_modes

        self.dtypes = FLOAT_DTYPES

        self._preprocessor_kwargs = {
            "sample_name": sample_name,
            "feature_name": feature_name,
            "with_std": False,
            "check_nans": check_nans,
        }

        # Define analysis-relevant meta data
        self.attrs = {"model": "BaseCrossModel"}
        self.attrs.update(
            {
                "software": "xeofs",
                "version": __version__,
                "date": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            }
        )

        # Initialize the data container only to avoid type errors
        # The actual data container will be initialized in respective subclasses
        # self.data: _BaseCrossModelDataContainer = _BaseCrossModelDataContainer()
        self.data = {}

    def _validate_data(self, views: Sequence[DataArray]):
        if not all(
            data[self.sample_name].size == views[0][self.sample_name].size
            for data in views
        ):
            raise ValueError("All views must have the same number of samples")
        if not all(data.ndim == 2 for data in views):
            raise ValueError("All views must have 2 dimensions")
        if not all(data.dtype in self.dtypes for data in views):
            raise ValueError("All views must have dtype of {}.".format(self.dtypes))
        if not all(data[self.feature_name].size >= self.n_modes for data in views):
            raise ValueError(
                "All views must have at least {} features.".format(self.n_modes)
            )

    def _process_init_pca_modes(self, n_modes):
        err_msg = "init_pca_modes must be either a float <= 1.0 or an integer > 1"
        n_modes_list = []
        n_modes_max = [
            min(self.n_samples_, n_features) for n_features in self.n_features_
        ]
        for n, n_max in zip(n_modes, n_modes_max):
            if isinstance(n, float):
                if n > 1.0:
                    raise ValueError(err_msg)
                n = int(n * n_max)
                n_modes_list.append(n)
            elif isinstance(n, int):
                if n <= 1:
                    raise ValueError(err_msg)
                n_modes_list.append(n)
            else:
                raise ValueError(err_msg)
        return n_modes_list

    def fit(
        self,
        views: Sequence[DataObject],
        dim: Hashable | Sequence[Hashable],
    ) -> Self:
        self.n_views_ = len(views)
        self.use_coslat = _process_parameter(
            "use_coslat", self.use_coslat, False, self.n_views_
        )
        self.init_pca_modes = _process_parameter(
            "init_pca_modes", self.init_pca_modes, 0.75, self.n_views_
        )

        # Preprocess the input data
        self.preprocessors = [
            Preprocessor(with_coslat=self.use_coslat[i], **self._preprocessor_kwargs)
            for i in range(self.n_views_)
        ]
        views2D: list[DataArray] = [
            preprocessor.fit_transform(data, dim)
            for preprocessor, data in zip(self.preprocessors, views)
        ]
        self._validate_data(views2D)
        self.n_features_ = [data.coords[self.feature_name].size for data in views2D]
        self.n_samples_ = views2D[0][self.sample_name].size

        self.data["input_data"] = views2D
        views2D = self._process_data(views2D)
        self.data["pca_data"] = views2D

        self._fit_algorithm(views2D)

        return self

    def _process_data(self, views: DataList) -> DataList:
        if self.pca:
            views = self._apply_pca(views)
        return views

    def _apply_pca(self, views: DataList):
        self.pca_models = []

        n_pca_modes = self._process_init_pca_modes(self.init_pca_modes)

        view_transformed = []

        for i, view in enumerate(views):
            # NOTE: coslat weighting already happens in Preprocessor class
            pca = EOF(n_modes=n_pca_modes[i], compute=self.compute)
            pca.fit(view, dim=self.sample_name)
            if self.compute:
                pca.compute()
            self.pca_models.append(pca)

            # TODO: method to get cumulative explained variance
            cum_exp_var_ratio = pca.explained_variance_ratio().cumsum()
            # Ensure that the sum of the explained variance ratio is always less than 1
            # Due to rounding errors the total sum may be slightly larger than 1,
            # which we counter by a small correction
            cum_exp_var_ratio -= 1e-6
            max_exp_var_ratio = cum_exp_var_ratio.isel(mode=-1).item()
            if (
                max_exp_var_ratio <= self.variance_fraction
                and max_exp_var_ratio <= 0.9999
            ):
                print(
                    "Warning: variance fraction {:.4f} is not reached. ".format(
                        self.variance_fraction
                    )
                    + "Only {:.4f} of variance is explained.".format(
                        cum_exp_var_ratio.isel(mode=-1).item()
                    )
                )
            n_modes_keep = (
                cum_exp_var_ratio.where(
                    cum_exp_var_ratio <= self.variance_fraction, drop=True
                ).size
                + 1
            )
            # Take at least 2 modes
            n_modes_keep = max(n_modes_keep, 2)

            scores = pca.scores(normalized=False).isel(mode=slice(0, n_modes_keep))
            scores = scores.rename({"mode": self.feature_name}).transpose(
                self.sample_name, self.feature_name
            )
            view_transformed.append(scores)
        return view_transformed

    @abstractmethod
    def _fit_algorithm(self, views: list[DataArray]) -> Self:
        raise NotImplementedError


class CCA(CCABaseModel):
    r"""Canonical Correlation Analysis.
    
    Canonical Correlation Analysis (CCA) identifies linear combinations of variables from multiple datasets that 
    maximize their mutual correlations. An optional regularisation parameter (ridge regression)
    can be used to improve the conditioning of the covariance matrix.

    The objective function of (regularised) CCA is:

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2  \}\\

        \text{subject to:}

        (1-c_1)w_1^TX_1^TX_1w_1+c_1w_1^Tw_1=n

        (1-c_2)w_2^TX_2^TX_2w_2+c_2w_2^Tw_2=n

    where :math:`c_i` are the regularization parameters for dataset.

    Parameters
    ----------
    n_modes : int, optional
        Number of latent dimensions to use, by default 10
    use_coslat : bool, optional
        Whether to use the square root of the cosine of the latitude as weights, by default False
    pca : bool, optional
        Whether to perform PCA on the input data, by default True
    variance_fraction : float, optional
        Fraction of variance to keep when performing PCA, by default 0.99
    init_pca_modes : int | float, optional
        Number of PCA modes to compute. If float, the number of modes is given by the fraction of maximum number of modes for the given data.
        A value of 1.0 will perform a full SVD of the data. Choosing a smaller value can increase computation speed. Default 0.75
    c : Sequence[float] | float], optional
        Regularisation parameter, by default 0 (no regularization)
    compute : bool, optional
        Whether to compute the decomposition immediately, by default True


    Notes
    -----
    This implementation is largely based on the MCCA class from the cca_zoo repository [3]_ .
    

    References
    ----------
    .. [1] Vinod, Hrishikesh _D. "Canonical ridge and econometrics of joint production." Journal of econometrics 4.2 (1976): 147-166.
    .. [2] Hotelling, Harold. "Relations between two sets of variates." Breakthroughs in statistics. Springer, New York, NY, 1992. 162-190.
    .. [3] Chapman et al., (2021). CCA-Zoo: A collection of Regularized, Deep Learning based, Kernel, and Probabilistic CCA methods in a scikit-learn style framework. Journal of Open Source Software, 6(68), 3823

    Examples
    --------
    >>> from xe.cross import CCA
    >>> model = CCA(n_modes=5)
    >>> model.fit(data)
    >>> can_loadings = model.canonical_loadings()

    """

    def __init__(
        self,
        n_modes: int = 2,
        use_coslat: bool = False,
        check_nans: bool = True,
        c: float = 0,
        pca: bool = True,
        variance_fraction: float = 0.99,
        init_pca_modes: float = 0.75,
        compute: bool = True,
        eps: float = 1e-6,
    ):
        super().__init__(
            n_modes=n_modes,
            use_coslat=use_coslat,
            check_nans=check_nans,
            pca=pca,
            compute=compute,
            variance_fraction=variance_fraction,
            init_pca_modes=init_pca_modes,
        )
        self.attrs.update({"model": "CCA"})
        self.c = c
        self.eps = eps

    def _fit_algorithm(self, views: list[DataArray]) -> Self:
        # Check input data
        [assert_not_complex(view) for view in views]

        self.c = _process_parameter("c", self.c, 0, self.n_views_)
        eigvals, eigvecs = self._solve_gevp(views)
        self.eigvals = eigvals
        self.eigvecs = eigvecs
        # Compute the weights for each view
        self._weights(eigvals, eigvecs, views)
        # Compute loadings (= normalized weights)
        self.data["loadings"] = [
            wght / self._apply_norm(wght, [self.feature_name])
            for wght in self.data["weights"]
        ]
        canonical_variates = self._transform(self.data["input_data"])
        self.data["variates"] = canonical_variates

        self.data["canonical_loadings"] = [
            xr.dot(data, vari, dims=self.sample_name, optimize=True)
            for data, vari in zip(self.data["input_data"], canonical_variates)
        ]

        # Compute explained variance
        # Transform the views using the loadings
        transformed_views = [
            xr.dot(view, loading, dims=self.feature_name)
            for view, loading in zip(self.data["input_data"], self.data["loadings"])
        ]
        # Calculate the variance of each latent dimension in the transformed views
        self.data["explained_variance"] = [
            transformed.var(self.sample_name) for transformed in transformed_views
        ]

        # Explained variance ratio
        self.data["total_variance"] = [
            view.var(self.sample_name, ddof=1).sum() for view in views
        ]

        # Calculate the explained variance ratio for each latent dimension for each view
        self.data["explained_variance_ratio"] = [
            exp_var / total_var
            for exp_var, total_var in zip(
                self.data["explained_variance"], self.data["total_variance"]
            )
        ]

        # Explained Covariance
        k = self.n_modes
        explained_covariance = []

        # just take the kth column of each transformed view and _compute_covariance
        for i in range(k):
            transformed_views_k = [
                view.isel(mode=slice(i, i + 1)) for view in transformed_views
            ]
            cov_ = self._apply_compute_covariance(
                transformed_views_k, dims_in=["sample", "mode"]
            )
            svals = self._compute_singular_values(cov_, dims_in=["mode1", "mode2"])
            explained_covariance.append(svals.isel(mode=0).item())
        self.data["explained_covariance"] = xr.DataArray(
            explained_covariance, dims=["mode"], coords={"mode": range(1, k + 1)}
        )

        minimum_dimension = min([view[self.feature_name].size for view in views])

        cov = self._apply_compute_covariance(views, dims_in=["sample", "feature"])
        S = self._compute_singular_values(cov, dims_in=["feature1", "feature2"])
        # select every other element starting from the first until the minimum dimension
        self.data["total_explained_covariance"] = (
            S.isel(mode=slice(0, None, 2)).isel(mode=slice(0, minimum_dimension)).sum()
        )
        self.data["explained_covariance_ratio"] = (
            self.data["explained_covariance"] / self.data["total_explained_covariance"]
        )

        return self

    def _compute_singular_values(
        self, x, dims_in=["feature1", "feature2"], dims_out=["mode"]
    ):
        svals = xr.apply_ufunc(
            np.linalg.svd,
            x,
            input_core_dims=[dims_in],
            output_core_dims=[dims_out],
            kwargs={"compute_uv": False},
            vectorize=False,
            dask="allowed",
        )
        svals = svals.assign_coords({"mode": range(1, svals.mode.size + 1)})
        return svals

    def _apply_norm(self, x, dims):
        return xr.apply_ufunc(
            np.linalg.norm,
            x,
            input_core_dims=[dims],
            output_core_dims=[[]],
            kwargs={"axis": -1},
            vectorize=True,
            dask="allowed",
        )

    def _solve_gevp(self, views: Sequence[DataArray], y=None, **kwargs):
        # Setup the eigenvalue problem
        C = self._C(views, dims_in=[self.sample_name, self.feature_name])
        D = self._D(views, **kwargs)
        self.splits = np.cumsum([view.shape[1] for view in views])
        # Solve the eigenvalue problem
        # Get the dimension of _C
        p = C.shape[0]
        subset_by_index = [p - self.n_modes, p - 1]
        # Solve the generalized eigenvalue problem Cx=lambda Dx using a subset of eigenvalues and eigenvectors
        [eigvals, eigvecs] = self._apply_eigh(C, D, subset_by_index=subset_by_index)
        # Sort the eigenvalues and eigenvectors in descending order
        idx_sorted_modes = eigvals.compute().argsort()[::-1]
        idx_sorted_modes = idx_sorted_modes.assign_coords(
            {"mode": range(idx_sorted_modes.mode.size)}
        )
        eigvals = eigvals.isel(mode=idx_sorted_modes)
        eigvecs = eigvecs.isel(mode=idx_sorted_modes).real
        # Set coordiantes
        coords_mode = range(1, eigvals.mode.size + 1)
        coords_feature = C.coords[self.feature_name + "1"].values
        eigvals = eigvals.assign_coords({"mode": coords_mode})
        eigvecs = eigvecs.assign_coords(
            {
                "mode": coords_mode,
                self.feature_name: coords_feature,
            }
        )
        return eigvals, eigvecs

    def _weights(self, eigvals, eigvecs, views, **kwargs):
        # split eigvecs into weights for each view
        # add 0 before the np ndarray splits
        idx = np.concatenate([[0], self.splits])
        self.data["weights"] = [
            eigvecs.isel({self.feature_name: slice(idx[i], idx[i + 1])})
            for i in range(len(idx) - 1)
        ]
        if self.pca:
            # go from weights in PCA space to weights in original space
            n_modes = [data.feature.size for data in self.data["pca_data"]]
            self.data["weights"] = [
                xr.dot(
                    pca.components()
                    .isel(mode=slice(0, n_modes[i]))
                    .rename({"mode": "temp_dim"}),
                    self.data["weights"][i].rename({"feature": "temp_dim"}),
                    dims="temp_dim",
                    optimize=True,
                )
                for i, pca in enumerate(self.pca_models)
            ]

    def _apply_eigh(self, a, b, subset_by_index):
        return xr.apply_ufunc(
            eigh,
            a,
            b,
            input_core_dims=[
                [self.feature_name + "1", self.feature_name + "2"],
                [self.feature_name + "1", self.feature_name + "2"],
            ],
            output_core_dims=[["mode"], ["feature", "mode"]],
            kwargs={"subset_by_index": subset_by_index},
            vectorize=False,
            dask="allowed",
        )

    def _C(self, views, dims_in):
        C = self._apply_compute_covariance(views, dims_in=dims_in)
        return C / len(views)

    def _apply_compute_covariance(
        self, views: Sequence[DataArray], dims_in, dims_out=None
    ) -> DataArray:
        if dims_out is None:
            dims_out = [dims_in[1] + "1", dims_in[1] + "2"]
        all_views = xr.concat(views, dim=dims_in[1])
        C = self._apply_cov(all_views, dims_in=dims_in, dims_out=dims_out)
        Ci = [
            self._apply_cov(view, dims_in=dims_in, dims_out=dims_out) for view in views
        ]
        return C - self._block_diag_dask(Ci, dims_in=dims_out)

    def _apply_cov(
        self, x, dims_in=["sample", "feature"], dims_out=["feature1", "feature2"]
    ):
        if x[dims_in[1]].size == 1:
            return xr.apply_ufunc(
                np.cov,
                x,
                input_core_dims=[dims_in],
                output_core_dims=[[]],
                kwargs={"rowvar": False},
                vectorize=False,
                dask="allowed",
            )
        else:
            C = xr.apply_ufunc(
                np.cov,
                x,
                input_core_dims=[dims_in],
                output_core_dims=[dims_out],
                kwargs={"rowvar": False},
                vectorize=False,
                dask="allowed",
            )
            feature_coords = x.coords[dims_in[1]].values
            C = C.assign_coords(
                {dims_out[0]: feature_coords, dims_out[1]: feature_coords}
            )
            return C

    def _block_diag_dask(self, views, dims_in=["feature1", "featur2"], dims_out=None):
        if dims_out is None:
            dims_out = dims_in
        if all(view.size == 1 for view in views):
            result = da.diag(np.array([view.item() for view in views]))
        else:
            # Extract underlying Dask arrays
            arrays = [da.asarray(view.data) for view in views]

            # Construct a block-diagonal dask array
            blocks = [
                [
                    darr2 if j == i else da.zeros((darr2.shape[0], darr1.shape[0]))
                    for j, darr1 in enumerate(views)
                ]
                for i, darr2 in enumerate(arrays)
            ]

            # Use Dask's block to stack the arrays
            blocked_array = da.block(blocks)

            # Convert the result back to a DataArray
            feature_coords = xr.concat(views, dim=dims_in[0])[dims_in[0]].values
            result = xr.DataArray(
                blocked_array,
                dims=dims_out,
                coords={dims_out[0]: feature_coords, dims_out[1]: feature_coords},
            )
        if any(isinstance(view.data, da.Array) for view in views):
            return result
        else:
            return result.compute()

    def _D(self, views):
        if self.pca:
            blocks = []
            for i, view in enumerate(views):
                pc = self.pca_models[i]
                feature_coords = view.coords[self.feature_name]
                n_features = feature_coords.size
                expvar = pc.explained_variance().isel(mode=slice(0, n_features))
                block = xr.DataArray(
                    da.diag((1 - self.c[i]) * expvar.data + self.c[i]),
                    dims=[self.feature_name + "1", self.feature_name + "2"],
                    coords={
                        self.feature_name + "1": feature_coords.values,
                        self.feature_name + "2": feature_coords.values,
                    },
                )
                block = block.compute()
                blocks.append(block)

        else:
            blocks = [self._apply_E(view, c) for view, c in zip(views, self.c)]

        D = self._block_diag_dask(blocks, dims_in=["feature1", "feature2"])

        D_smallest_eig = self._apply_smallest_eigval(D, dims=["feature1", "feature2"])
        D_smallest_eig = D_smallest_eig - self.eps
        identity_matrix = xr.DataArray(np.eye(D.shape[0]), dims=D.dims, coords=D.coords)
        D = D - D_smallest_eig * identity_matrix
        return D / len(views)

    def _apply_E(self, view, c):
        E = xr.apply_ufunc(
            self._E,
            view,
            input_core_dims=[[self.sample_name, self.feature_name]],
            output_core_dims=[[self.feature_name + "1", self.feature_name + "2"]],
            kwargs={"c": c},
            vectorize=False,
            dask="allowed",
        )
        feature_coords = view.coords[self.feature_name].values
        E = E.assign_coords(
            {
                self.feature_name + "1": feature_coords,
                self.feature_name + "2": feature_coords,
            }
        )
        return E

    def _E(self, view, c):
        return (1 - c) * np.cov(view, rowvar=False) + c * np.eye(view.shape[1])

    def _apply_smallest_eigval(self, D, dims):
        return xr.apply_ufunc(
            self._smallest_eigval,
            D,
            input_core_dims=[dims],
            output_core_dims=[[]],
            vectorize=True,
            dask="allowed",
        )

    def _smallest_eigval(self, D):
        return min(0, np.linalg.eigvalsh(D).min())

    def weights(self) -> list[DataObject]:
        weights = [
            prep.inverse_transform_components(wghts)
            for prep, wghts in zip(self.preprocessors, self.data["weights"])
        ]
        return weights

    def _transform(self, views: Sequence[DataArray]) -> list[DataArray]:
        transformed_views = []
        for i, view in enumerate(views):
            transformed_view = xr.dot(view, self.data["weights"][i], dims="feature")
            transformed_views.append(transformed_view)
        return transformed_views

    def transform(self, views: Sequence[DataObject]) -> list[DataArray]:
        """Transform the input data into the canonical space.

        Parameters
        ----------
        views : list[DataArray | Dataset]
            Input data to transform

        """
        view_preprocessed = []
        for i, view in enumerate(views):
            view_preprocessed = self.preprocessors[i].transform(view)

        transformed_views = self._transform(view_preprocessed)

        unstacked_transformed_views = []
        for i, view in enumerate(transformed_views):
            unstacked_view = self.preprocessors[i].inverse_transform_scores(view)
            unstacked_transformed_views.append(unstacked_view)
        return unstacked_transformed_views

    def components(self, normalize: bool = True) -> list[DataObject]:
        """Get the canonical loadings for each view."""
        can_loads = self.data["canonical_loadings"]
        input_data = self.data["input_data"]
        variates = self.data["variates"]

        if normalize:
            # Compute correlations
            loadings = [
                (
                    loads
                    / data[self.sample_name].size
                    / data.std(self.sample_name)
                    / vari.std(self.sample_name)
                ).clip(-1, 1)
                for loads, data, vari in zip(can_loads, input_data, variates)
            ]
        else:
            loadings = can_loads

        loadings = [
            prep.inverse_transform_components(load)
            for prep, load in zip(self.preprocessors, loadings)
        ]
        return loadings

    def scores(self) -> list[DataArray]:
        """Get the canonical variates for each view."""
        variates = []
        for i, view in enumerate(self.data["variates"]):
            vari = self.preprocessors[i].inverse_transform_scores(view)
            variates.append(vari)
        return variates

    def explained_variance(self) -> list[DataArray]:
        """Get the explained variance for each view."""
        return self.data["explained_variance"]

    def explained_variance_ratio(self) -> list[DataArray]:
        """Get the explained variance ratio for each view."""
        return self.data["explained_variance_ratio"]

    def explained_covariance(self) -> DataArray:
        """Get the explained covariance."""
        return self.data["explained_covariance"]

    def explained_covariance_ratio(self) -> DataArray:
        """Get the explained covariance ratio."""
        return self.data["explained_covariance_ratio"]



import numpy as np
import pytest
import xarray as xr

from xeofs.data_container import DataContainer
from xeofs.single import EOF, EOFRotator

from ...utilities import data_is_dask, skip_if_missing_engine


@pytest.fixture
def eof_model(mock_data_array, dim):
    eof = EOF(n_modes=5)
    eof.fit(mock_data_array, dim)
    return eof


@pytest.fixture
def eof_model_delayed(mock_dask_data_array, dim):
    eof = EOF(n_modes=5, compute=False, check_nans=False)
    eof.fit(mock_dask_data_array, dim)
    return eof


def test_init():
    # Instantiate the EOFRotator class
    eof_rotator = EOFRotator(n_modes=3, power=2, max_iter=100, rtol=1e-6)

    assert eof_rotator._params["n_modes"] == 3
    assert eof_rotator._params["power"] == 2
    assert eof_rotator._params["max_iter"] == 100
    assert eof_rotator._params["rtol"] == 1e-6


@pytest.mark.parametrize(
    "dim",
    [
        ("time"),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit(eof_model):
    eof_rotator = EOFRotator(n_modes=3)
    eof_rotator.fit(eof_model)

    assert hasattr(
        eof_rotator, "model_data"
    ), 'The attribute "model_data" should be populated after fitting.'
    assert hasattr(
        eof_rotator, "data"
    ), 'The attribute "data" should be populated after fitting.'
    assert isinstance(eof_rotator.model_data, DataContainer)
    assert isinstance(eof_rotator.data, DataContainer)


@pytest.mark.parametrize(
    "dim",
    [
        ("time"),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_transform(eof_model, mock_data_array):
    eof_rotator = EOFRotator(n_modes=3)
    eof_rotator.fit(eof_model)
    projections = eof_rotator.transform(mock_data_array)

    assert isinstance(projections, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        ("time"),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_inverse_transform(eof_model):
    eof_rotator = EOFRotator(n_modes=3)
    eof_rotator.fit(eof_model)
    scores = eof_rotator.data["scores"].sel(mode=1)
    Xrec = eof_rotator.inverse_transform(scores)

    assert isinstance(Xrec, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        ("time"),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_explained_variance(eof_model):
    eof_rotator = EOFRotator(n_modes=3)
    eof_rotator.fit(eof_model)
    exp_var = eof_rotator.explained_variance()
    exp_var_ref = eof_model.explained_variance().sel(mode=slice(1, 3))

    assert isinstance(exp_var, xr.DataArray)
    # 3 modes should be returned
    assert exp_var.size == 3
    # The explained variance should be positive
    assert (exp_var > 0).all()
    # The sum of the explained variance should be the same
    # before and after rotation
    xr.testing.assert_allclose(exp_var.sum(), exp_var_ref.sum())


@pytest.mark.parametrize(
    "dim",
    [
        ("time"),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_explained_variance_ratio(eof_model):
    eof_rotator = EOFRotator(n_modes=3)
    eof_rotator.fit(eof_model)
    exp_var_ratio = eof_rotator.explained_variance_ratio()
    exp_var_ratio_ref = eof_model.explained_variance_ratio().sel(mode=slice(1, 3))

    assert isinstance(exp_var_ratio, xr.DataArray)
    # 3 modes should be returned
    assert exp_var_ratio.size == 3
    # The explained variance should be positive
    assert (exp_var_ratio > 0).all()
    # The total of the explained variance ratio should be <= 1
    assert exp_var_ratio.sum() <= 1
    # The sum of the explained variance should be the same
    # before and after rotation
    xr.testing.assert_allclose(exp_var_ratio.sum(), exp_var_ratio_ref.sum())


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components(eof_model):
    eof_rotator = EOFRotator(n_modes=3)
    eof_rotator.fit(eof_model)
    components = eof_rotator.components()

    assert isinstance(components, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores(eof_model):
    eof_rotator = EOFRotator(n_modes=3)
    eof_rotator.fit(eof_model)
    scores = eof_rotator.scores()

    assert isinstance(scores, xr.DataArray)


@pytest.mark.parametrize(
    "dim, compute",
    [
        (("time",), True),
        (("lat", "lon"), True),
        (("lon", "lat"), True),
        (("time",), False),
        (("lat", "lon"), False),
        (("lon", "lat"), False),
    ],
)
def test_compute(eof_model_delayed, compute):
    eof_rotator = EOFRotator(n_modes=5, compute=compute, max_iter=20, rtol=1e-4)
    eof_rotator.fit(eof_model_delayed)

    if compute:
        assert not data_is_dask(eof_rotator.data["explained_variance"])
        assert not data_is_dask(eof_rotator.data["components"])
        assert not data_is_dask(eof_rotator.data["rotation_matrix"])

    else:
        assert data_is_dask(eof_rotator.data["explained_variance"])
        assert data_is_dask(eof_rotator.data["components"])
        assert data_is_dask(eof_rotator.data["rotation_matrix"])


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
@pytest.mark.parametrize("engine", ["h5netcdf", "netcdf4", "zarr"])
def test_save_load(dim, mock_data_array, tmp_path, engine):
    """Test save/load methods in EOF class, ensuring that we can
    roundtrip the model and get the same results when transforming
    data."""
    skip_if_missing_engine(engine)

    original_unrotated = EOF()
    original_unrotated.fit(mock_data_array, dim)

    original = EOFRotator()
    original.fit(original_unrotated)

    # Save the EOF model
    original.save(tmp_path / "eof", engine=engine)

    # Check that the EOF model has been saved
    assert (tmp_path / "eof").exists()

    # Recreate the model from saved file
    loaded = EOFRotator.load(tmp_path / "eof", engine=engine)

    # Check that the params and DataContainer objects match
    assert original.get_params() == loaded.get_params()
    assert all([key in loaded.data for key in original.data])
    for key in original.data:
        if original.data._allow_compute[key]:
            assert loaded.data[key].equals(original.data[key])
        else:
            # but ensure that input data is not saved by default
            assert loaded.data[key].size <= 1
            assert loaded.data[key].attrs["placeholder"] is True

    # Test that the recreated model can be used to transform new data
    assert np.allclose(
        original.transform(mock_data_array), loaded.transform(mock_data_array)
    )

    # The loaded model should also be able to inverse_transform new data
    assert np.allclose(
        original.inverse_transform(original.scores()),
        loaded.inverse_transform(loaded.scores()),
    )


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_serialize_deserialize_dataarray(dim, mock_data_array):
    """Test roundtrip serialization when the model is fit on a DataArray."""
    model = EOF()
    model.fit(mock_data_array, dim)
    rotator = EOFRotator()
    rotator.fit(model)
    dt = rotator.serialize()
    rebuilt_rotator = EOFRotator.deserialize(dt)
    assert np.allclose(
        rotator.transform(mock_data_array), rebuilt_rotator.transform(mock_data_array)
    )


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_serialize_deserialize_dataset(dim, mock_dataset):
    """Test roundtrip serialization when the model is fit on a Dataset."""
    model = EOF()
    model.fit(mock_dataset, dim)
    rotator = EOFRotator()
    rotator.fit(model)
    dt = rotator.serialize()
    rebuilt_rotator = EOFRotator.deserialize(dt)
    assert np.allclose(
        rotator.transform(mock_dataset), rebuilt_rotator.transform(mock_dataset)
    )



import dask
import numpy as np
import xarray as xr
from typing_extensions import Self

from ..utils.data_types import DataArray, DataVar, DataVarBound, Dims
from ..utils.xarray_utils import compute_sqrt_cos_lat_weights, feature_ones_like
from .transformer import Transformer


class Scaler(Transformer):
    """Scale the data along sample dimensions.

    Scaling includes (i) removing the mean and, optionally, (ii) dividing by the standard deviation,
    (iii) multiplying by the square root of cosine of latitude weights (area weighting; coslat weighting),
    and (iv) multiplying by additional user-defined weights.

    Parameters
    ----------
    with_center : bool, default=True
        If True, the data is centered by subtracting the mean.
    with_std : bool, default=True
        If True, the data is divided by the standard deviation.
    with_coslat : bool, default=False
        If True, the data is multiplied by the square root of cosine of latitude weights.
    weights : DataArray | Dataset, optional
        Weights to be applied to the data. Must have the same dimensions as the data.
        If None, no weights are applied.
    """

    def __init__(
        self,
        with_center: bool = True,
        with_std: bool = False,
        with_coslat: bool = False,
        compute: bool = True,
    ):
        super().__init__()
        self.with_center = with_center
        self.with_std = with_std
        self.with_coslat = with_coslat
        self.compute = compute

        self.mean_ = xr.DataArray(name="mean_")
        self.std_ = xr.DataArray(name="std_")
        self.coslat_weights_ = xr.DataArray(name="coslat_weights_")
        self.weights_ = xr.DataArray(name="weights_")

    def get_serialization_attrs(self) -> dict:
        return dict(
            mean_=self.mean_,
            std_=self.std_,
            coslat_weights_=self.coslat_weights_,
            weights_=self.weights_,
        )

    def _verify_input(self, X, name: str):
        if not isinstance(X, (xr.DataArray, xr.Dataset)):
            raise TypeError(f"{name} must be an xarray DataArray or Dataset")

    def _process_weights(self, X: DataVarBound, weights) -> DataVarBound:
        if weights is None:
            wghts: DataVarBound = feature_ones_like(X, self.feature_dims)
        else:
            wghts: DataVarBound = weights

        return wghts

    def fit(
        self,
        X: DataVar,
        sample_dims: Dims,
        feature_dims: Dims,
        weights: DataVar | None = None,
    ) -> Self:
        """Fit the scaler to the data.

        Parameters
        ----------
        X : DataArray | Dataset
            Data to be scaled.
        sample_dims : sequence of hashable
            Dimensions along which the data is considered to be a sample.
        feature_dims : sequence of hashable
            Dimensions along which the data is considered to be a feature.
        weights : DataArray | Dataset, optional
            Weights to be applied to the data. Must have the same dimensions as the data.
            If None, no weights are applied.

        """
        # Check input types
        self._verify_input(X, "data")

        self.sample_dims = sample_dims
        self.feature_dims = feature_dims
        # Store sample and feature dimensions for later use
        self.dims = {"sample": sample_dims, "feature": feature_dims}

        params = self.get_params()

        # Scaling parameters are computed along sample dimensions
        if params["with_center"]:
            self.mean_: DataVar = X.mean(self.sample_dims)

        if params["with_std"]:
            self.std_: DataVar = X.std(self.sample_dims).clip(
                min=np.finfo(np.float32).eps
            )

        if params["with_coslat"]:
            self.coslat_weights_: DataVar = compute_sqrt_cos_lat_weights(
                data=X, feature_dims=self.feature_dims
            )

        # Convert None weights to ones
        self.weights_: DataVar = self._process_weights(X, weights)

        if self.get_params()["compute"]:
            (self.mean_, self.std_, self.coslat_weights_, self.weights_) = dask.compute(
                self.mean_,
                self.std_,
                self.coslat_weights_,
                self.weights_,
            )

        return self

    def transform(self, X: DataVarBound) -> DataVarBound:
        """Scale the data.

        Parameters
        ----------
        data : DataArray | Dataset
            Data to be scaled.

        Returns
        -------
        DataArray | Dataset
            Scaled data.

        """
        self._verify_input(X, "X")

        params = self.get_params()

        if params["with_center"]:
            X = X - self.mean_
        if params["with_std"]:
            X = X / self.std_
        if params["with_coslat"]:
            X = X * self.coslat_weights_

        X = X * self.weights_
        return X

    def fit_transform(
        self,
        X: DataVarBound,
        sample_dims: Dims,
        feature_dims: Dims,
        weights: DataVarBound | None = None,
    ) -> DataVarBound:
        return self.fit(X, sample_dims, feature_dims, weights).transform(X)

    def inverse_transform_data(self, X: DataVarBound) -> DataVarBound:
        """Unscale the data.

        Parameters
        ----------
        X : DataArray | DataSet
            Data to be unscaled.

        Returns
        -------
        DataArray | DataSet
            Unscaled data.

        """
        self._verify_input(X, "X")

        params = self.get_params()
        X = X / self.weights_
        if params["with_coslat"]:
            X = X / self.coslat_weights_
        if params["with_std"]:
            X = X * self.std_
        if params["with_center"]:
            X = X + self.mean_

        return X

    def inverse_transform_components(self, X: DataVarBound) -> DataVarBound:
        return X

    def inverse_transform_scores(self, X: DataArray) -> DataArray:
        return X

    def inverse_transform_scores_unseen(self, X: DataArray) -> DataArray:
        return X



import numpy as np
import pytest
from dask.array import Array as DaskArray  # type: ignore

from xeofs.linalg.decomposer import Decomposer

from ..utilities import data_is_dask


def compute_max_exp_var(singular_values, data):
    """Compute the maximal cumulative explained variance by all components."""

    total_variance = data.var("sample", ddof=1).sum("feature")
    explained_variance = singular_values**2 / (data.sample.size - 1)
    explained_variance_ratio = explained_variance / total_variance
    explained_variance_ratio_cumsum = explained_variance_ratio.cumsum("mode")
    return explained_variance_ratio_cumsum.isel(mode=-1).item()


@pytest.fixture
def decomposer():
    return Decomposer(n_modes=2, random_state=42)


@pytest.fixture
def mock_data_array(mock_data_array):
    data2d = mock_data_array.stack(sample=("time",), feature=("lat", "lon")).dropna(
        "feature"
    )
    return data2d - data2d.mean("sample")


@pytest.fixture
def mock_dask_data_array(mock_data_array):
    return mock_data_array.chunk({"sample": 2})


@pytest.fixture
def mock_complex_data_array(mock_data_array):
    return mock_data_array * (1 + 1j)


@pytest.fixture
def test_complex_dask_data_array(mock_complex_data_array):
    return mock_complex_data_array.chunk({"sample": 2})


def test_init(decomposer):
    assert decomposer.n_modes == 2
    assert decomposer.random_state == 42


def test_fit_full(mock_data_array):
    decomposer = Decomposer(n_modes=2, solver="full")
    decomposer.fit(mock_data_array)
    assert "U_" in decomposer.__dict__
    assert "s_" in decomposer.__dict__
    assert "V_" in decomposer.__dict__

    # Check that indeed 2 modes are returned
    assert decomposer.U_.shape[1] == 2
    assert decomposer.s_.shape[0] == 2
    assert decomposer.V_.shape[1] == 2


def test_fit_full_matrices(mock_data_array):
    decomposer = Decomposer(
        n_modes=2, solver="full", solver_kwargs={"full_matrices": False}
    )
    decomposer.fit(mock_data_array)
    assert "U_" in decomposer.__dict__
    assert "s_" in decomposer.__dict__
    assert "V_" in decomposer.__dict__

    # Check that indeed 2 modes are returned
    assert decomposer.U_.shape[1] == 2
    assert decomposer.s_.shape[0] == 2
    assert decomposer.V_.shape[1] == 2


def test_fit_randomized(mock_data_array):
    decomposer = Decomposer(n_modes=2, solver="randomized", random_state=42)
    decomposer.fit(mock_data_array)
    assert "U_" in decomposer.__dict__
    assert "s_" in decomposer.__dict__
    assert "V_" in decomposer.__dict__

    # Check that indeed 2 modes are returned
    assert decomposer.U_.shape[1] == 2
    assert decomposer.s_.shape[0] == 2
    assert decomposer.V_.shape[1] == 2


def test_fit_dask_full(mock_dask_data_array):
    # The Dask SVD solver has no parameter 'random_state' but 'seed' instead,
    # so let's create a new decomposer for this case
    decomposer = Decomposer(n_modes=2, solver="full")
    decomposer.fit(mock_dask_data_array)
    assert "U_" in decomposer.__dict__
    assert "s_" in decomposer.__dict__
    assert "V_" in decomposer.__dict__

    # Check if the Dask SVD solver has been used
    assert isinstance(decomposer.U_.data, DaskArray)
    assert isinstance(decomposer.s_.data, DaskArray)
    assert isinstance(decomposer.V_.data, DaskArray)

    # Check that indeed 2 modes are returned
    assert decomposer.U_.shape[1] == 2
    assert decomposer.s_.shape[0] == 2
    assert decomposer.V_.shape[1] == 2


@pytest.mark.parametrize("compute", [True, False])
def test_fit_dask_randomized(mock_dask_data_array, compute):
    # The Dask SVD solver has no parameter 'random_state' but 'seed' instead,
    # so this should be automatically converted depending on the solver
    decomposer = Decomposer(
        n_modes=2, solver="randomized", compute=compute, random_state=42
    )
    decomposer.fit(mock_dask_data_array)
    assert "U_" in decomposer.__dict__
    assert "s_" in decomposer.__dict__
    assert "V_" in decomposer.__dict__

    # Check that indeed 2 modes are returned
    assert decomposer.U_.shape[1] == 2
    assert decomposer.s_.shape[0] == 2
    assert decomposer.V_.shape[1] == 2

    is_dask_before = data_is_dask(mock_dask_data_array)
    U_is_dask_after = data_is_dask(decomposer.U_)
    s_is_dask_after = data_is_dask(decomposer.s_)
    V_is_dask_after = data_is_dask(decomposer.V_)
    # Check if the Dask SVD solver has been used
    assert is_dask_before
    if compute:
        assert not U_is_dask_after
        assert not s_is_dask_after
        assert not V_is_dask_after
    else:
        assert U_is_dask_after
        assert s_is_dask_after
        assert V_is_dask_after


def test_fit_complex(mock_complex_data_array):
    decomposer = Decomposer(n_modes=2, solver="randomized", random_state=42)
    decomposer.fit(mock_complex_data_array)
    assert "U_" in decomposer.__dict__
    assert "s_" in decomposer.__dict__
    assert "V_" in decomposer.__dict__

    # Check that indeed 2 modes are returned
    assert decomposer.U_.shape[1] == 2
    assert decomposer.s_.shape[0] == 2
    assert decomposer.V_.shape[1] == 2

    # Check that U and V are complex
    assert np.iscomplexobj(decomposer.U_.data)
    assert np.iscomplexobj(decomposer.V_.data)


@pytest.mark.parametrize(
    "data",
    ["real", "complex", "dask_real"],
)
def test_random_state(
    data, mock_data_array, mock_complex_data_array, mock_dask_data_array
):
    match data:
        case "real":
            X = mock_data_array
        case "complex":
            X = mock_complex_data_array
        case "dask_real":
            X = mock_dask_data_array
        case _:
            raise ValueError(f"Unrecognized data type '{data}'.")

    decomposer = Decomposer(
        n_modes=2, solver="randomized", random_state=42, compute=True
    )
    decomposer.fit(X)
    U1 = decomposer.U_.data

    # Refit
    decomposer.fit(X)
    U2 = decomposer.U_.data

    # Check that the results are the same
    assert np.all(U1 == U2)


@pytest.mark.parametrize(
    "target_variance, solver",
    [
        (0.1, "randomized"),
        (0.5, "randomized"),
        (0.9, "randomized"),
        (0.99, "randomized"),
        (0.1, "full"),
        (0.5, "full"),
        (0.9, "full"),
        (0.99, "full"),
    ],
)
def test_decompose_via_variance_threshold(mock_data_array, target_variance, solver):
    """Test that the decomposer returns the correct number of modes to explain the target variance."""
    decomposer = Decomposer(
        n_modes=target_variance, solver=solver, init_rank_reduction=0.9
    )
    decomposer.fit(mock_data_array)
    s = decomposer.s_

    # Compute total variance and test whether variance threshold is reached
    max_explained_variance_ratio = compute_max_exp_var(s, mock_data_array)
    assert (
        max_explained_variance_ratio >= target_variance
    ), f"Expected >= {target_variance:.2f}, got {max_explained_variance_ratio:2f}"

    # We still get a truncated version of the SVD
    assert s.mode.size < min(mock_data_array.shape)


def test_raise_warning_for_low_init_rank_reduction(mock_data_array):
    target_variance = 0.5
    init_rank_reduction = 0.1
    decomposer = Decomposer(
        n_modes=target_variance, init_rank_reduction=init_rank_reduction
    )
    warn_msg = "Dataset has .* components, explaining .* of the variance. However, .*explained variance was requested. Please consider increasing `init_rank_reduction`"
    with pytest.warns(UserWarning, match=warn_msg):
        decomposer.fit(mock_data_array)


def test_compute_at_least_one_component(mock_data_array):
    """"""
    target_variance = 0.5
    init_rank_reduction = 0.01
    decomposer = Decomposer(
        n_modes=target_variance, init_rank_reduction=init_rank_reduction
    )

    # Warning is raised to indicate that the value of init_rank_reduction is too low
    warn_msg = "`init_rank_reduction=.*` is too low resulting in zero components. One component will be computed instead."
    with pytest.warns(UserWarning, match=warn_msg):
        decomposer.fit(mock_data_array)

    # At least one mode is computed
    s = decomposer.s_
    assert s.mode.size >= 1


@pytest.mark.parametrize(
    "solver",
    ["full", "randomized"],
)
def test_dask_array_based_on_target_variance(mock_dask_data_array, solver):
    target_variance = 0.5
    decomposer = Decomposer(
        n_modes=target_variance, init_rank_reduction=0.9, solver=solver, compute=False
    )

    err_msg = "Estimating the number of modes to keep based on variance is not supported with dask arrays.*"
    with pytest.raises(ValueError, match=err_msg):
        assert data_is_dask(mock_dask_data_array)
        decomposer.fit(mock_dask_data_array)



import scipy as sc
import xarray as xr
from statsmodels.stats.multitest import multipletests as statsmodels_multipletests

from ..constants import MULTIPLE_TESTS


def pearson_correlation(
    data1,
    data2,
    correction=None,
    alpha=0.05,
    sample_name="sample",
    feature_name="feature",
):
    """Compute Pearson correlation between two xarray objects.

    Additionally, compute two-sided p-values and adjust them for multiple testing.

    Parameters
    ----------
    data1 : xr.DataArray
         First data array.
    data2 : xr.DataArray
        Second data array.
    correction : Optional, str
        Method used for testing and adjustment of pvalues. Can be
            either the full name or initial letters.  Available methods are:
            - bonferroni : one-step correction
            - sidak : one-step correction
            - holm-sidak : step down method using Sidak adjustments
            - holm : step-down method using Bonferroni adjustments
            - simes-hochberg : step-up method (independent)
            - hommel : closed method based on Simes tests (non-negative)
            - fdr_bh : Benjamini/Hochberg (non-negative) (default)
            - fdr_by : Benjamini/Yekutieli (negative)
            - fdr_tsbh : two stage fdr correction (non-negative)
            - fdr_tsbky : two stage fdr correction (non-negative)
    alpha : float
        The desired family-wise error rate.

    Returns
    -------
    corr : xr.DataArray
        Pearson correlation between the original data and the scores.
    pvalues : xr.DataArray
        Adjusted p-values for the Pearson correlation.

    """

    def _correlation_coefficients_numpy(X, Y):
        """Compute Pearson correlation coefficients assuming centered data."""
        X = X / X.std(0)
        Y = Y / Y.std(0)
        return X.conj().T @ Y / X.shape[0]

    n_samples = data1.shape[0]

    # Compute Pearson correlation coefficients
    sample_name_x = "sample_dim_x"
    sample_name_y = "sample_dim_y"
    data1 = data1.rename({sample_name: sample_name_x})
    data2 = data2.rename({sample_name: sample_name_y})
    feature_name_x = data1.dims[1]
    feature_name_y = data2.dims[1]
    corr = xr.apply_ufunc(
        _correlation_coefficients_numpy,
        data1,
        data2,
        input_core_dims=[
            [sample_name_x, feature_name_x],
            [sample_name_y, feature_name_y],
        ],
        output_core_dims=[[feature_name_x, feature_name_y]],
        dask="allowed",
    )

    # Compute two-sided p-values
    pvals = _compute_pvalues(corr, n_samples)

    if correction is not None:
        # Adjust p-values for multiple testing
        rejected, pvals = _multipletests(
            pvals, dim=feature_name, alpha=alpha, method=correction
        )
        return corr, pvals

    else:
        return corr, pvals


def _compute_pvalues(pearsonr, n_samples: int) -> xr.DataArray:
    # Compute two-sided p-values
    # Reference: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.pearsonr.html#r8c6348c62346-1
    a = n_samples / 2 - 1
    dist = sc.stats.beta(a, a, loc=-1, scale=2)  # type: ignore
    pvals = 2 * xr.apply_ufunc(
        dist.cdf,
        -abs(pearsonr),
        input_core_dims=[[]],
        output_core_dims=[[]],
        dask="allowed",
        vectorize=False,
    )
    return pvals


def _multipletests(p, dim, alpha=0.05, method=None, **multipletests_kwargs):
    """Apply statsmodels.stats.multitest.multipletests for 2-dimensional
    xr.objects.

    Parameters
    ----------
    p: xr.object
        uncorrected p-values.
    alpha: (optional float)
        FWER, family-wise error rate. Defaults to 0.05.
    method: str
        Method used for testing and adjustment of pvalues. Can be
            either the full name or initial letters.  Available methods are:
            - bonferroni : one-step correction
            - sidak : one-step correction
            - holm-sidak : step down method using Sidak adjustments
            - holm : step-down method using Bonferroni adjustments
            - simes-hochberg : step-up method (independent)
            - hommel : closed method based on Simes tests (non-negative)
            - fdr_bh : Benjamini/Hochberg (non-negative)
            - fdr_by : Benjamini/Yekutieli (negative)
            - fdr_tsbh : two stage fdr correction (non-negative)
            - fdr_tsbky : two stage fdr correction (non-negative)
    **multipletests_kwargs (optional dict): is_sorted, returnsorted
        see statsmodels.stats.multitest.multitest

    Returns:
        reject (xr.object): true for hypothesis that can be rejected for given
            alpha
        pvals_corrected (xr.object): p-values corrected for multiple tests

    """
    if method is None:
        raise ValueError(
            f"Please indicate a method using the 'method=...' keyword. "
            f"Select from {MULTIPLE_TESTS}"
        )
    elif method not in MULTIPLE_TESTS:
        raise ValueError(
            f"Your method '{method}' is not in the accepted methods: {MULTIPLE_TESTS}"
        )

    rej, pvals_corr, *_ = xr.apply_ufunc(
        statsmodels_multipletests,
        p,
        input_core_dims=[[dim]],
        output_core_dims=[[dim], [dim], [], []],
        dask="allowed",
        vectorize=True,
    )
    return rej, pvals_corr



from typing import Sequence

import numpy as np
import xarray as xr
from typing_extensions import Self

from ..base_model import BaseModel
from ..data_container import DataContainer
from ..linalg.rotation import promax
from ..preprocessing import PCA, Preprocessor, Whitener
from ..utils.data_types import DataArray, DataObject
from ..utils.xarray_utils import argsort_dask, get_deterministic_sign_multiplier
from .cpcca import CPCCA, ComplexCPCCA, HilbertCPCCA


class CPCCARotator(CPCCA):
    """Rotate a solution obtained from ``xe.cross.CPCCA``.

    Rotate the obtained components and scores of a CPCCA model to increase
    interpretability. The algorithm here is based on the approach of Cheng &
    Dunkerton (1995) [1]_ and adapted to the CPCCA framework [2]_.

    Parameters
    ----------
    n_modes : int, default=10
        Specify the number of modes to be rotated.
    power : int, default=1
        Set the power for the Promax rotation. A ``power`` value of 1 results in
        a Varimax rotation.
    max_iter : int or None, default=None
        Determine the maximum number of iterations for the computation of the
        rotation matrix. If not specified, defaults to 1000 if ``compute=True``
        and 100 if ``compute=False``, since we can't terminate a lazy
        computation based using ``rtol``.
    rtol : float, default=1e-8
        Define the relative tolerance required to achieve convergence and
        terminate the iterative process.
    compute : bool, default=True
        Whether to compute the rotation immediately.

    References
    ----------
    .. [1] Cheng, X. & Dunkerton, T. J. Orthogonal Rotation of Spatial Patterns
        Derived from Singular Value Decomposition Analysis. J. Climate 8,
        2631–2643 (1995).
    .. [2] Swenson, E. Continuum Power CCA: A Unified Approach for Isolating
        Coupled Modes. Journal of Climate 28, 1016–1030 (2015).

    Examples
    --------

    Perform a CPCCA analysis:

    >>> model = CPCCA(n_modes=10)
    >>> model.fit(X, Y, dim='time')

    Then, apply varimax rotation to first 5 components and scores:

    >>> rotator = CPCCARotator(n_modes=5)
    >>> rotator.fit(model)

    Retrieve the rotated components and scores:

    >>> rotator.components()
    >>> rotator.scores()

    """

    def __init__(
        self,
        n_modes: int = 10,
        power: int = 1,
        max_iter: int | None = None,
        rtol: float = 1e-8,
        compute: bool = True,
    ):
        BaseModel.__init__(self)

        if max_iter is None:
            max_iter = 1000 if compute else 100

        # Define model parameters
        self._params = {
            "n_modes": n_modes,
            "power": power,
            "max_iter": max_iter,
            "rtol": rtol,
            "compute": compute,
        }

        # Define analysis-relevant meta data
        self.attrs.update({"model": "Rotated CPCCA"})
        self.attrs.update(self._params)

        # Attach empty objects
        self.preprocessor1 = Preprocessor()
        self.preprocessor2 = Preprocessor()
        self.pca1 = PCA()
        self.pca2 = PCA()
        self.whitener1 = Whitener()
        self.whitener2 = Whitener()
        self.data = DataContainer()
        self.model_data = DataContainer()

        self.sorted = False

    def get_serialization_attrs(self) -> dict:
        return dict(
            data=self.data,
            model_data=self.model_data,
            preprocessor1=self.preprocessor1,
            preprocessor2=self.preprocessor2,
            pca1=self.pca1,
            pca2=self.pca2,
            whitener1=self.whitener1,
            whitener2=self.whitener2,
            sorted=self.sorted,
            sample_name=self.sample_name,
            feature_name=self.feature_name,
        )

    def _fit_algorithm(self, model) -> Self:
        self.preprocessor1 = model.preprocessor1
        self.preprocessor2 = model.preprocessor2
        self.pca1 = model.pca1
        self.pca2 = model.pca2
        self.whitener1 = model.whitener1
        self.whitener2 = model.whitener2
        self.sample_name = model.sample_name
        self.feature_name = model.feature_name
        self.sorted = False

        common_feature_dim = "common_feature_dim"
        feature_name = model.feature_name

        n_modes = self._params["n_modes"]
        power = self._params["power"]
        max_iter = self._params["max_iter"]
        rtol = self._params["rtol"]

        # Construct the combined vector of loadings
        # NOTE: In the methodology
        # used by Cheng & Dunkerton (CD95), the combined vectors are "loaded" or
        # weighted with the square root of the singular values, akin to what is
        # done in standard Varimax rotation. While this approach ensures that
        # the resulting projections are still uncorrelated as in the unrotated
        # solution, it does not conserve the squared covariance fraction, i.e.
        # the amount of explained squared covariance can be different before and
        # after rotation. The authors then introduced a so-called "covariance
        # fraction" which is conserved under rotation, but does not have a clear
        # interpretation as the term covariance fraction is only correct when
        # both data sets X and Y are equal and MCA reduces to PCA.
        svalues = model.data["singular_values"].sel(mode=slice(1, n_modes))
        scaling = np.sqrt(svalues)

        # Get unrotated singular vectors
        Qx = model.data["components1"].sel(mode=slice(1, n_modes))
        Qy = model.data["components2"].sel(mode=slice(1, n_modes))

        # Unwhiten and back-transform into physical space
        Qx = self.whitener1.inverse_transform_components(Qx)
        Qy = self.whitener2.inverse_transform_components(Qy)
        Qx = self.pca1.inverse_transform_components(Qx)
        Qy = self.pca2.inverse_transform_components(Qy)

        # Rename the feature dimension to a common name so that the combined vectors can be concatenated
        Qx = Qx.rename({feature_name[0]: common_feature_dim})
        Qy = Qy.rename({feature_name[1]: common_feature_dim})

        loadings = xr.concat([Qx, Qy], dim=common_feature_dim) * scaling

        # Rotate loadings
        promax_kwargs = {"power": power, "max_iter": max_iter, "rtol": rtol}
        rot_loadings, rot_matrix, phi_matrix = promax(
            loadings=loadings,
            feature_dim=common_feature_dim,
            compute=self._params["compute"],
            **promax_kwargs,
        )

        # Assign coordinates to the rotation/correlation matrices
        rot_matrix = rot_matrix.assign_coords(
            mode_m=np.arange(1, rot_matrix.mode_m.size + 1),
            mode_n=np.arange(1, rot_matrix.mode_n.size + 1),
        )
        phi_matrix = phi_matrix.assign_coords(
            mode_m=np.arange(1, phi_matrix.mode_m.size + 1),
            mode_n=np.arange(1, phi_matrix.mode_n.size + 1),
        )

        # Rotated (loaded) singular vectors
        Qx_rot = rot_loadings.isel(
            {common_feature_dim: slice(0, Qx.coords[common_feature_dim].size)}
        )
        Qy_rot = rot_loadings.isel(
            {common_feature_dim: slice(Qx.coords[common_feature_dim].size, None)}
        )

        # Rename the common feature dimension to the original feature names
        Qx_rot = Qx_rot.rename({common_feature_dim: feature_name[0]})
        Qy_rot = Qy_rot.rename({common_feature_dim: feature_name[1]})

        # For consistency with the unrotated model classes, we transform the pattern vectors
        # into the whitened PC space
        Qx_rot = self.pca1.transform_components(Qx_rot)
        Qy_rot = self.pca2.transform_components(Qy_rot)
        Qx_rot = self.whitener1.transform_components(Qx_rot)
        Qy_rot = self.whitener2.transform_components(Qy_rot)

        # Normalization factor of singular vectors
        norm1_rot = xr.apply_ufunc(
            np.linalg.norm,
            Qx_rot,
            input_core_dims=[[feature_name[0], "mode"]],
            output_core_dims=[["mode"]],
            exclude_dims={feature_name[0]},
            kwargs={"axis": 0},
            vectorize=False,
            dask="allowed",
        )
        norm2_rot = xr.apply_ufunc(
            np.linalg.norm,
            Qy_rot,
            input_core_dims=[[feature_name[1], "mode"]],
            output_core_dims=[["mode"]],
            exclude_dims={feature_name[1]},
            kwargs={"axis": 0},
            vectorize=False,
            dask="allowed",
        )

        # Rotated (normalized) singular vectors
        Qx_rot = Qx_rot / norm1_rot
        Qy_rot = Qy_rot / norm2_rot

        # CD95 call the quantity "norm1 * norm2" the "explained covariance"
        explained_covariance = norm1_rot * norm2_rot
        squared_covariance = explained_covariance**2

        # Reorder according to squared covariance
        idx_modes_sorted = argsort_dask(squared_covariance, "mode")[::-1]  # type: ignore
        idx_modes_sorted.coords.update(squared_covariance.coords)

        # Rotate scores using rotation matrix
        scores1 = model.data["scores1"].sel(mode=slice(1, n_modes))
        scores2 = model.data["scores2"].sel(mode=slice(1, n_modes))

        scores1 = self.whitener1.inverse_transform_scores(scores1)
        scores2 = self.whitener2.inverse_transform_scores(scores2)

        # Normalize scores
        scores1 = scores1 / scaling
        scores2 = scores2 / scaling

        RinvT = self._compute_rot_mat_inv_trans(
            rot_matrix, input_dims=("mode_m", "mode_n")
        )
        # Rename dimension mode to ensure that dot product has dimensions (sample x mode) as output
        scores1 = scores1.rename({"mode": "mode_m"})
        scores2 = scores2.rename({"mode": "mode_m"})
        RinvT = RinvT.rename({"mode_n": "mode"})
        scores1_rot = xr.dot(scores1, RinvT, dims="mode_m") * norm1_rot
        scores2_rot = xr.dot(scores2, RinvT, dims="mode_m") * norm2_rot

        # Ensure consitent signs for deterministic output
        modes_sign = get_deterministic_sign_multiplier(rot_loadings, common_feature_dim)
        Qx_rot = Qx_rot * modes_sign
        Qy_rot = Qy_rot * modes_sign
        scores1_rot = scores1_rot * modes_sign
        scores2_rot = scores2_rot * modes_sign

        # Create data container for Rotator and original model data
        self.model_data.add(name="singular_values", data=model.data["singular_values"])
        self.model_data.add(name="components1", data=model.data["components1"])
        self.model_data.add(name="components2", data=model.data["components2"])

        # Assigning input data to the Rotator object allows us to inherit some functionalities from the original model
        # like squared_covariance_fraction(), homogeneous_patterns() etc.
        self.data.add(
            name="input_data1", data=model.data["input_data1"], allow_compute=False
        )
        self.data.add(
            name="input_data2", data=model.data["input_data2"], allow_compute=False
        )
        self.data.add(name="components1", data=Qx_rot)
        self.data.add(name="components2", data=Qy_rot)
        self.data.add(name="scores1", data=scores1_rot)
        self.data.add(name="scores2", data=scores2_rot)
        self.data.add(name="squared_covariance", data=squared_covariance)
        self.data.add(
            name="total_squared_covariance",
            data=model.data["total_squared_covariance"],
        )

        self.data.add(name="idx_modes_sorted", data=idx_modes_sorted)
        self.data.add(name="norm1", data=norm1_rot)
        self.data.add(name="norm2", data=norm2_rot)
        self.data.add(name="rotation_matrix", data=rot_matrix)
        self.data.add(name="phi_matrix", data=phi_matrix)
        self.data.add(name="modes_sign", data=modes_sign)

        # Assign analysis-relevant meta data
        self.data.set_attrs(self.attrs)

        return self

    def fit(self, model: CPCCA) -> Self:
        """Rotate the solution obtained from ``xe.cross.CPCCA``.

        Parameters
        ----------
        model : ``xe.cross.CPCCA``
            The CPCCA model to be rotated.

        """
        self._fit_algorithm(model)

        if self._params["compute"]:
            self.compute()

        return self

    def transform(
        self,
        X: DataObject | None = None,
        Y: DataObject | None = None,
        normalized: bool = False,
    ) -> DataArray | list[DataArray]:
        """Transform the data.

        Parameters
        ----------
        X, Y: DataObject | None
            Data to be transformed. At least one of them must be provided.
        normalized: bool, default=False
            Whether to return L2 normalized scores.

        Returns
        -------
        Sequence[DataArray] | DataArray
            Transformed data.

        """
        # raise error if no data is provided
        if X is None and Y is None:
            raise ValueError("No data provided. Please provide X and/or Y.")

        n_modes = self._params["n_modes"]
        rot_matrix = self.data["rotation_matrix"]
        RinvT = self._compute_rot_mat_inv_trans(
            rot_matrix, input_dims=("mode_m", "mode_n")
        )
        RinvT = RinvT.rename({"mode_n": "mode"})

        scaling = self.model_data["singular_values"].sel(mode=slice(1, n_modes))
        scaling = np.sqrt(scaling)

        results = []

        if X is not None:
            # Select the (non-rotated) singular vectors of the first dataset
            comps1 = self.model_data["components1"].sel(mode=slice(1, n_modes))

            # Preprocess the data
            comps1 = self.whitener1.inverse_transform_components(comps1)
            comps1 = self.pca1.inverse_transform_components(comps1)
            X = self.preprocessor1.transform(X)

            # Compute non-rotated scores by projecting the data onto non-rotated components
            projections1 = xr.dot(X, comps1) / scaling
            # Rotate the scores
            projections1 = projections1.rename({"mode": "mode_m"})
            projections1 = xr.dot(projections1, RinvT, dims="mode_m")
            # Reorder according to variance
            if self.sorted:
                projections1 = projections1.isel(
                    mode=self.data["idx_modes_sorted"].values
                ).assign_coords(mode=projections1.mode)
            # Adapt the sign of the scores
            projections1 = projections1 * self.data["modes_sign"]

            # Unscale the scores
            if not normalized:
                projections1 = projections1 * self.data["norm1"]

            # Unstack the projections
            projections1 = self.preprocessor1.inverse_transform_scores(projections1)

            results.append(projections1)

        if Y is not None:
            # Select the (non-rotated) singular vectors of the second dataset
            comps2 = self.model_data["components2"].sel(mode=slice(1, n_modes))

            # Preprocess the data
            comps2 = self.whitener2.inverse_transform_components(comps2)
            comps2 = self.pca2.inverse_transform_components(comps2)
            Y = self.preprocessor2.transform(Y)

            # Compute non-rotated scores by project the data onto non-rotated components
            projections2 = xr.dot(Y, comps2) / scaling
            # Rotate the scores
            projections2 = projections2.rename({"mode": "mode_m"})
            projections2 = xr.dot(projections2, RinvT, dims="mode_m")
            # Reorder according to variance
            if self.sorted:
                projections2 = projections2.isel(
                    mode=self.data["idx_modes_sorted"].values
                ).assign_coords(mode=projections2.mode)
            # Determine the sign of the scores
            projections2 = projections2 * self.data["modes_sign"]

            # Unscale the scores
            if not normalized:
                projections2 = projections2 * self.data["norm2"]

            # Unstack the projections
            projections2 = self.preprocessor2.inverse_transform_scores(projections2)

            results.append(projections2)

        if len(results) == 0:
            raise ValueError("provide at least one of [`X`, `Y`]")
        elif len(results) == 1:
            return results[0]
        else:
            return results

    def _post_compute(self):
        """Leave sorting until after compute because it can't be done lazily."""
        self._sort_by_variance()

    def _sort_by_variance(self):
        """Re-sort the mode dimension of all data variables by variance explained."""
        if not self.sorted:
            for key in self.data.keys():
                if "mode" in self.data[key].dims and key != "idx_modes_sorted":
                    self.data[key] = (
                        self.data[key]
                        .isel(mode=self.data["idx_modes_sorted"].values)
                        .assign_coords(mode=self.data[key].mode)
                    )
        self.sorted = True

    def _compute_rot_mat_inv_trans(self, rotation_matrix, input_dims) -> xr.DataArray:
        """Compute the inverse transpose of the rotation matrix.

        For orthogonal rotations (e.g., Varimax), the inverse transpose is equivalent
        to the rotation matrix itself. For oblique rotations (e.g., Promax), the simplification
        does not hold.

        Returns
        -------
        rotation_matrix : xr.DataArray

        """
        if self._params["power"] > 1:
            # inverse matrix
            rotation_matrix = xr.apply_ufunc(
                np.linalg.inv,
                rotation_matrix,
                input_core_dims=[(input_dims)],
                output_core_dims=[(input_dims[::-1])],
                vectorize=False,
                dask="allowed",
            )
            # transpose matrix
            rotation_matrix = rotation_matrix.conj().transpose(*input_dims)
        return rotation_matrix


class ComplexCPCCARotator(CPCCARotator, ComplexCPCCA):
    """Rotate a solution obtained from ``xe.cross.ComplexCPCCA``.

    Rotate the obtained components and scores of a CPCCA model to increase
    interpretability. The algorithm here is based on the approach of Cheng &
    Dunkerton (1995) [1]_ and adapted to the CPCCA framework [2]_.


    Parameters
    ----------
    n_modes : int, default=10
        Specify the number of modes to be rotated.
    power : int, default=1
        Set the power for the Promax rotation. A ``power`` value of 1 results in
        a Varimax rotation.
    max_iter : int, default=1000
        Determine the maximum number of iterations for the computation of the
        rotation matrix.
    rtol : float, default=1e-8
        Define the relative tolerance required to achieve convergence and
        terminate the iterative process.
    squared_loadings : bool, default=False
        Specify the method for constructing the combined vectors of loadings. If
        True, the combined vectors are loaded with the singular values (termed
        "squared loadings"), conserving the squared covariance under rotation.
        This allows estimation of mode importance after rotation. If False, the
        combined vectors are loaded with the square root of the singular values,
        following the method described by Cheng & Dunkerton.
    compute: bool, default=True
        Whether to compute the rotation immediately.

    References
    ----------
    .. [1] Cheng, X. & Dunkerton, T. J. Orthogonal Rotation of Spatial Patterns
        Derived from Singular Value Decomposition Analysis. J. Climate 8,
        2631–2643 (1995).
    .. [3] Swenson, E. Continuum Power CCA: A Unified Approach for Isolating
        Coupled Modes. Journal of Climate 28, 1016–1030 (2015).

    Examples
    --------

    Perform a CPCCA analysis:

    >>> model = ComplexCPCCA(n_modes=10)
    >>> model.fit(X, Y, dim='time')

    Then, apply varimax rotation to first 5 components and scores:

    >>> rotator = ComplexCPCCARotator(n_modes=5)
    >>> rotator.fit(model)

    Retrieve the rotated components and scores:

    >>> rotator.components()
    >>> rotator.scores()

    """

    def __init__(self, **kwargs):
        CPCCARotator.__init__(self, **kwargs)
        self.attrs.update({"model": "Rotated Complex CPCCA"})


class HilbertCPCCARotator(ComplexCPCCARotator, HilbertCPCCA):
    """Rotate a solution obtained from ``xe.cross.HilbertCPCCA``.

    Rotate the obtained components and scores of a CPCCA model to increase
    interpretability. The algorithm here is based on the approach of Cheng &
    Dunkerton (1995) [1]_ and adapted to the CPCCA framework [2]_.


    Parameters
    ----------
    n_modes : int, default=10
        Specify the number of modes to be rotated.
    power : int, default=1
        Set the power for the Promax rotation. A ``power`` value of 1 results in
        a Varimax rotation.
    max_iter : int, default=1000
        Determine the maximum number of iterations for the computation of the
        rotation matrix.
    rtol : float, default=1e-8
        Define the relative tolerance required to achieve convergence and
        terminate the iterative process.
    squared_loadings : bool, default=False
        Specify the method for constructing the combined vectors of loadings. If
        True, the combined vectors are loaded with the singular values (termed
        "squared loadings"), conserving the squared covariance under rotation.
        This allows estimation of mode importance after rotation. If False, the
        combined vectors are loaded with the square root of the singular values,
        following the method described by Cheng & Dunkerton.
    compute: bool, default=True
        Whether to compute the rotation immediately.

    References
    ----------
    .. [1] Cheng, X. & Dunkerton, T. J. Orthogonal Rotation of Spatial Patterns
        Derived from Singular Value Decomposition Analysis. J. Climate 8,
        2631–2643 (1995).
    .. [3] Swenson, E. Continuum Power CCA: A Unified Approach for Isolating
        Coupled Modes. Journal of Climate 28, 1016–1030 (2015).

    Examples
    --------

    Perform a CPCCA analysis:

    >>> model = HilbertCPCCA(n_modes=10)
    >>> model.fit(X, Y, dim='time')

    Then, apply varimax rotation to first 5 components and scores:

    >>> rotator = HilbertCPCCARotator(n_modes=5)
    >>> rotator.fit(model)

    Retrieve the rotated components and scores:

    >>> rotator.components()
    >>> rotator.scores()

    """

    def __init__(self, **kwargs):
        ComplexCPCCARotator.__init__(self, **kwargs)
        self.attrs.update({"model": "Rotated Hilbert CPCCA"})

    def transform(
        self, X: DataObject | None = None, Y: DataObject | None = None, normalized=False
    ) -> Sequence[DataArray]:
        """Transform the data."""
        # Here we make use of the Method Resolution Order (MRO) to call the
        # transform method of the first class in the MRO after `CPCCARotator`
        # that has a transform method. In this case it will be `HilbertCPCCA`,
        # which will raise an error because it does not have a transform method.
        return super(CPCCARotator, self).transform(X, Y, normalized)



import numpy as np

from ._svd import _SVD


def _fractional_matrix_power(C, power, **kwargs):
    """Compute the fractional matrix power of a symmetric matrix using SVD.

    Note: This function is a simplified version of the fractional_matrix_power
    function from the scipy library. However, the scipy function does not
    support dask arrays due to the use of np.asarray.
    """
    if C.shape[0] != C.shape[1]:
        raise ValueError("Matrix must be square.")

    svd = _SVD(n_modes="all", **kwargs)
    _, s, V = svd.fit_transform(C)

    # cut off small singular values
    is_above_zero = s > np.finfo(s.dtype).eps
    V = V[:, is_above_zero]
    s = s[is_above_zero]

    # TODO: use hermitian=True for numpy>=2.0
    # V, s, _ = np.linalg.svd(C, hermitian=True)
    C_scaled = V @ np.diag(s**power) @ V.conj().T

    # Even if the input matrix is real and symmetric, the output matrix might
    # be complex due to numerical errors. In this case, we return the real part
    if np.iscomplexobj(C):
        return C_scaled
    else:
        return C_scaled.real



"""
Weighted EOF analysis
========================

Weighted EOF analysis (in S-mode) maximises the temporal variance
considering each gridpoint with a different weight. We compare the
results for an EOF analysis based on (1) the covariance matrix, (2) area
weighting based on latitude (coslat weighting), (3) the correlation matrix
and finally (4) correlation matrix + coslat weighting.

Load packages and data:
"""

import matplotlib.pyplot as plt
import seaborn as sns
import xarray as xr
from cartopy.crs import Orthographic, PlateCarree
from matplotlib.gridspec import GridSpec

import xeofs as xe

sns.set_context("paper")

t2m = xr.tutorial.load_dataset("air_temperature")["air"]

# %%
# Perform the actual analysis

components = []
scores = []
# (1) Based on covariance matrix
model_cov = xe.single.EOF(n_modes=5, standardize=False, use_coslat=False)
model_cov.fit(t2m, "time")
components.append(model_cov.components())
scores.append(model_cov.scores())
# (2) Based on coslat weighted covariance matrix
model_lat = xe.single.EOF(n_modes=5, standardize=False, use_coslat=True)
model_lat.fit(t2m, "time")
components.append(model_lat.components())
scores.append(model_lat.scores())
# (3) Based on correlation matrix
model_cor = xe.single.EOF(n_modes=5, standardize=True, use_coslat=False)
model_cor.fit(t2m, "time")
components.append(model_cor.components())
scores.append(model_cor.scores())
# (4) Based on coslat weighted correlation matrix
model_cor_lat = xe.single.EOF(n_modes=5, standardize=True, use_coslat=True)
model_cor_lat.fit(t2m, "time")
components.append(model_cor_lat.components())
scores.append(model_cor_lat.scores())


# %%
# Create figure showing the first mode for all 4 cases

proj = Orthographic(central_latitude=30, central_longitude=-80)
kwargs = {
    "cmap": "mako",
    "transform": PlateCarree(),
    "vmin": 0,
}
titles = [
    "(1) Covariances",
    "(2) Covariances + coslat",
    "(3) Correlation",
    "(4) Correlation + coslat",
]
fig = plt.figure(figsize=(10, 12))
gs = GridSpec(4, 2)
ax_pcs = [fig.add_subplot(gs[i, 0]) for i in range(4)]
ax_eofs = [fig.add_subplot(gs[i, 1], projection=proj) for i in range(4)]

for i, (a1, a2) in enumerate(zip(ax_eofs, ax_pcs)):
    a1.coastlines(color=".5")
    components[i].sel(mode=1).plot(ax=a1, **kwargs)
    scores[i].sel(mode=1).plot(ax=a2, color="darkred")
    a2.set_xlabel("")
    a1.set_title("", loc="center")
    a2.set_title("", loc="center")
    a2.set_title(titles[i], loc="left", weight="bold")
    if i < 3:
        a2.set_xticks([], [])
        sns.despine(ax=a2, trim=True, bottom=True)
    else:
        sns.despine(ax=a2, trim=True, bottom=False)

plt.tight_layout()
plt.savefig("weighted_eof.jpg", dpi=200)



# %%

import numpy as np
import xarray as xr
from typing_extensions import Self

from ..utils.data_types import DataArray, DataObject
from ..utils.sanity_checks import assert_not_complex
from ..utils.xarray_utils import get_matrix_rank
from ..utils.xarray_utils import total_variance as compute_total_variance
from ._numpy._sparse_pca import compute_rspca, compute_spca
from .base_model_single_set import BaseModelSingleSet


class SparsePCA(BaseModelSingleSet):
    """
    Sparse PCA via Variable Projection.

    Given a data matrix, Sparse PCA via Variable Projection [1]_ computes a set of sparse components that can optimally reconstruct the input data. The amount of sparsity is controlled by the L1 penalty coefficient, specified by the parameter `alpha`. Additionally, ridge shrinkage can be applied to improve conditioning.

    This implementation uses randomized methods for linear algebra to accelerate computations. The quality of the approximation can be controlled via the oversampling parameter `oversample` and `n_subspace`, which specifies the number of subspace iterations.

    Variable Projection for Sparse PCA solves the following optimization problem:

        minimize :math:`\\frac{1}{2} \\| X - X B A^T \\|^2 + \\alpha \\|B\\|_1 + \\frac{1}{2} \\beta \\|B\\|^2`

    Parameters
    ----------
    n_modes : int, default=2
        Number of modes to calculate.
    alpha : float, default=0.001
        Sparsity controlling parameter. Higher values lead to sparser components.
    beta : float, default=0.001
        Amount of ridge shrinkage to apply in order to improve conditioning.
    regularizer : {'l0', 'l1'}, default='l1'
        Type of sparsity-inducing regularizer. The L1 norm (also known as LASSO)
        leads to a soft-threshold operator (default). The L0 norm is implemented
        via a hard-threshold operator.
    max_iter : int, default=500
        Maximum number of iterations to perform before exiting.
    tol : float, default=1e-5
        Stopping tolerance for reconstruction error.
    robust : bool, default=False
        Use a robust algorithm to compute the sparse PCA.
    oversample : int, default=10
        Controls the oversampling of column space. Increasing this parameter
        may improve numerical accuracy.
    n_subspace : int, default=2
        Parameter to control the number of subspace iterations. Increasing this
        parameter may improve numerical accuracy.
    n_blocks : int, default=2
        Parameter to control how many blocks of columns the input matrix
        should be split into. A larger number requires less fast memory, but
        increases computational time.
    standardize : bool, default=False
        Whether to standardize the input data, i.e., each feature will have a variance of 1.
    use_coslat : bool, default=False
        Whether to use cosine of latitude for scaling.
    sample_name : str, default="sample"
        Name of the sample dimension.
    feature_name : str, default="feature"
        Name of the feature dimension.
    compute : bool, default=True
        Whether to compute elements of the model eagerly, or to defer computation.
        If True, the following steps will be computed sequentially:
        1) the preprocessor scaler,
        2) optional NaN checks,
        3) SVD decomposition,
        4) scores and components.
    random_state : int, optional
        Seed for the random number generator.
    solver : {"auto", "full", "randomized"}, default="randomized"
        Solver to use for the SVD computation.
    solver_kwargs : dict, default={}
        Additional keyword arguments to be passed to the SVD solver.

    References
    ----------
    .. [1] Erichson, N. B. et al. Sparse Principal Component Analysis via Variable Projection. SIAM J. Appl. Math. 80, 977-1002 (2020).

    Notes
    -----
    This implementation is adapted from the code provided by the authors of the paper [1]_, which is part of the Ristretto library (https://github.com/erichson/ristretto).

    Examples
    --------
    >>> model = xe.single.SparsePCA(n_modes=2, alpha=1e-4)
    >>> model.fit(data, "time")
    >>> components = model.components()
    """

    def __init__(
        self,
        n_modes: int = 2,
        alpha: float = 1e-3,
        beta: float = 1e-3,
        robust: bool = False,
        regularizer: str = "l1",
        max_iter: int = 500,
        tol: float = 1e-6,
        oversample: int = 10,
        n_subspace: int = 1,
        n_blocks: int = 1,
        center: bool = True,
        standardize: bool = False,
        use_coslat: bool = False,
        sample_name: str = "sample",
        feature_name: str = "feature",
        check_nans=True,
        compute: bool = True,
        random_state: int | None = None,
        solver: str = "auto",
        solver_kwargs: dict = {},
        **kwargs,
    ):
        super().__init__(
            n_modes=n_modes,
            center=center,
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            sample_name=sample_name,
            feature_name=feature_name,
            compute=compute,
            random_state=random_state,
            solver=solver,
            solver_kwargs=solver_kwargs,
            **kwargs,
        )
        self.attrs.update({"model": "Sparse PCA"})
        self._params.update(
            {
                "alpha": alpha,
                "beta": beta,
                "robust": robust,
                "regularizer": regularizer,
                "max_iter": max_iter,
                "tol": tol,
                "oversample": oversample,
                "n_subspace": n_subspace,
                "n_blocks": n_blocks,
            }
        )

    def _fit_algorithm(self, X: DataArray) -> Self:
        sample_name = self.sample_name
        feature_name = self.feature_name

        # Check if the data is real
        # NOTE: Complex data is not supported, it's likely possible but current numpy implementation
        # of sparse_pca needs to be adpated, mainly changing matrix transpose to conjugate transpose.
        # http://arxiv.org/abs/1804.00341
        assert_not_complex(X)

        # Compute the total variance
        total_variance = compute_total_variance(X, dim=sample_name)

        # Compute matrix rank
        rank = get_matrix_rank(X)

        # Decide whether to use exact or randomized algorithm
        is_small_data = max(X.shape) < 500
        solver = self._params["solver"]

        match solver:
            case "auto":
                use_exact = (
                    True if is_small_data and self.n_modes > int(0.8 * rank) else False
                )
            case "full":
                use_exact = True
            case "randomized":
                use_exact = False
            case _:
                raise ValueError(
                    f"Unrecognized solver '{solver}'. "
                    "Valid options are 'auto', 'full', and 'randomized'."
                )

        decomposing_kwargs = dict(
            n_components=self.n_modes,
            alpha=self._params["alpha"],
            beta=self._params["beta"],
            robust=self._params["robust"],
            regularizer=self._params["regularizer"],
            max_iter=self._params["max_iter"],
            tol=self._params["tol"],
            compute=self._params["compute"],
        )
        if use_exact:
            decomposing_algorithm = compute_spca
        else:
            decomposing_algorithm = compute_rspca
            decomposing_kwargs.update(
                dict(
                    oversample=self._params["oversample"],
                    n_subspace=self._params["n_subspace"],
                    n_blocks=self._params["n_blocks"],
                    random_state=self._params["random_state"],
                )
            )

        # Fit the data
        # We obtain the follwing outputs defined in Erichson et al. 2020
        # variable : notation used by Erichson et al.
        # components : sparse weight matrix B
        # components_normal : orthonormal matrix A
        # exp_var : eigenvalues
        components, components_normal, exp_var = xr.apply_ufunc(
            decomposing_algorithm,
            X,
            input_core_dims=[[sample_name, feature_name]],
            output_core_dims=[[feature_name, "mode"], [feature_name, "mode"], ["mode"]],
            dask="allowed",
            kwargs=decomposing_kwargs,
        )

        # Add coordinates to the results
        exp_var.name = "explained_variance"
        exp_var = exp_var.assign_coords({"mode": np.arange(1, self.n_modes + 1)})

        components.name = "sparse_weight_vectors"
        components = components.assign_coords(
            {
                feature_name: X.coords[feature_name],
                "mode": np.arange(1, self.n_modes + 1),
            },
        )

        components_normal.name = "orthonormal_weight_vectors"
        components_normal = components_normal.assign_coords(
            {
                feature_name: X.coords[feature_name],
                "mode": np.arange(1, self.n_modes + 1),
            },
        )

        # Transform the data
        scores = xr.dot(X, components, dims=feature_name)
        scores.name = "scores"

        norms = xr.apply_ufunc(
            np.linalg.norm,
            scores,
            input_core_dims=[[sample_name]],
            output_core_dims=[[]],
            exclude_dims=set((sample_name,)),
            kwargs={"axis": -1},
            dask="allowed",
            vectorize=False,
        )
        norms.name = "component_norms"

        # Store the results
        self.data.add(X, "input_data", allow_compute=False)
        self.data.add(components, "components")
        self.data.add(components_normal, "components_normal")
        self.data.add(scores, "scores")
        self.data.add(norms, "norms")
        self.data.add(exp_var, "explained_variance")
        self.data.add(total_variance, "total_variance")

        self.data.set_attrs(self.attrs)
        return self

    def _transform_algorithm(self, X: DataObject) -> DataArray:
        feature_name = self.preprocessor.feature_name

        components = self.data["components"]

        # Project the data
        projections = xr.dot(X, components, dims=feature_name)
        projections.name = "scores"

        return projections

    def _inverse_transform_algorithm(self, scores: DataArray) -> DataArray:
        # Reconstruct the data
        comps = self.data["components_normal"].sel(mode=scores.mode)

        reconstructed_data = xr.dot(comps.conj(), scores, dims="mode")
        reconstructed_data.name = "reconstructed_data"

        return reconstructed_data

    def components(self) -> DataObject:
        """Return the sparse components.

        The components are given by the sparse weight matrix B in the optimization problem.

        Returns
        -------
        components: DataArray | Dataset | list[DataArray]
            Components of the fitted model.

        """
        return super().components()

    def scores(self, normalized: bool = False) -> DataArray:
        """Return the component scores.

        The component scores :math:`U` are defined as the projection of the fitted
        data :math:`X` onto the sparse weight components :math:`B`.

        .. math::
            U = X B


        Parameters
        ----------
        normalized : bool, default=False
            Whether to normalize the scores by the L2 norm.

        Returns
        -------
        components: DataArray | Dataset | list[DataArray]
            Scores of the fitted model.

        """
        return super().scores(normalized=normalized)

    def explained_variance(self) -> DataArray:
        """Return the explained variance.


        Returns
        -------
        explained_variance: DataArray
            Explained variance.
        """
        return self.data["explained_variance"]

    def explained_variance_ratio(self) -> DataArray:
        """Return explained variance ratio.

        The explained variance ratio :math:`\\gamma_i` is the variance explained
        by each mode normalized by the total variance. It is defined as

        .. math::
            \\gamma_i = \\frac{\\lambda_i}{\\sum_{j=1}^M \\lambda_j}

        where :math:`\\lambda_i` is the explained variance of the :math:`i`-th mode and :math:`M` is the total number of modes.

        Returns
        -------
        explained_variance_ratio: DataArray
            Explained variance ratio.
        """
        exp_var_ratio = self.data["explained_variance"] / self.data["total_variance"]
        exp_var_ratio.attrs.update(self.data["explained_variance"].attrs)
        exp_var_ratio.name = "explained_variance_ratio"
        return exp_var_ratio



import numpy as np
import pytest
import xarray as xr

from xeofs.preprocessing import PCA

from ..utilities import (
    assert_expected_coords,
    assert_expected_dims,
    data_is_dask,
)

# =============================================================================
# GENERALLY VALID TEST CASES
# =============================================================================
N_SAMPLE_DIMS = [1]
N_FEATURE_DIMS = [1]
INDEX_POLICY = ["index"]
NAN_POLICY = ["no_nan"]
DASK_POLICY = ["no_dask", "dask"]
SEED = [0]

VALID_TEST_DATA = [
    (ns, nf, index, nan, dask)
    for ns in N_SAMPLE_DIMS
    for nf in N_FEATURE_DIMS
    for index in INDEX_POLICY
    for nan in NAN_POLICY
    for dask in DASK_POLICY
]


def generate_well_conditioned_data(lazy=False):
    t = np.linspace(0, 50, 200)
    std = 0.1
    X = np.sin(t)[:, None] + np.random.normal(0, std, size=(200, 3))
    X[:, 1] = X[:, 1] ** 3
    X[:, 2] = abs(X[:, 2]) ** (0.5)
    X = xr.DataArray(
        X,
        dims=["sample", "feature"],
        coords={"sample": np.arange(200), "feature": np.arange(3)},
        name="X",
    )
    X = X - X.mean("sample")
    if lazy:
        X = X.chunk({"sample": 5, "feature": -1})
    return X


# TESTS
# =============================================================================
@pytest.mark.parametrize("lazy", [False, True])
def test_fit(lazy):
    data = generate_well_conditioned_data(lazy)

    pca = PCA(n_modes=2)
    pca.fit(data)


@pytest.mark.parametrize("lazy", [False, True])
@pytest.mark.parametrize("use_pca", [True, False])
def test_transform(lazy, use_pca):
    data = generate_well_conditioned_data(lazy)

    pca = PCA(n_modes=2, use_pca=use_pca)
    pca.fit(data)

    # Transform data
    transformed_data = pca.transform(data)
    transformed_data2 = pca.transform(data)
    assert transformed_data.identical(transformed_data2)

    assert isinstance(transformed_data, xr.DataArray)
    assert transformed_data.ndim == 2
    assert transformed_data.dims == ("sample", "feature")

    # Consistent dask behaviour
    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(transformed_data)
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize("lazy", [False, True])
@pytest.mark.parametrize("use_pca", [True, False])
def test_fit_transform(lazy, use_pca):
    data = generate_well_conditioned_data(lazy)

    pca = PCA(n_modes=2, use_pca=use_pca)

    # Transform data
    transformed_data = pca.fit_transform(data)
    transformed_data2 = pca.transform(data)
    assert transformed_data.identical(transformed_data2)

    assert isinstance(transformed_data, xr.DataArray)
    assert transformed_data.ndim == 2
    assert transformed_data.dims == ("sample", "feature")

    # Consistent dask behaviour
    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(transformed_data)
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize("lazy", [False, True])
@pytest.mark.parametrize("use_pca", [True, False])
def test_invserse_transform_data(lazy, use_pca):
    data = generate_well_conditioned_data(lazy)

    pca = PCA(n_modes=2, use_pca=use_pca)
    pca.fit(data)

    transformed = pca.transform(data)
    untransformed = pca.inverse_transform_data(transformed)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(untransformed)

    # Unstacked data has dimensions of original data
    assert_expected_dims(data, untransformed, policy="all")
    # Unstacked data has coordinates of original data
    assert_expected_coords(data, untransformed, policy="all")
    # inverse transform should not change dask-ness
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize("n_modes", [1, 2, 3])
def test_transform_pca_n_modes(n_modes):
    data = generate_well_conditioned_data()

    pca = PCA(use_pca=True, n_modes=n_modes)
    transformed = pca.fit_transform(data)

    # PCA reduces dimensionality
    assert transformed.shape[1] == n_modes


@pytest.mark.parametrize("use_pca", [True, False])
def test_transform_keep_coordinates(use_pca):
    X = generate_well_conditioned_data()

    pca = PCA(use_pca=use_pca, n_modes="all")
    transformed = pca.fit_transform(X)

    assert len(transformed.coords) == len(X.coords)



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any

import numpy as np
import xarray as xr
from tqdm import trange

from .._version import __version__
from ..data_container import DataContainer
from ..single import EOF
from ..utils.data_types import DataArray


class _BaseBootstrapper(ABC):
    """Bootstrap a model to obtain significant modes and confidence intervals."""

    def __init__(self, n_bootstraps=20, seed=None):
        self._params = {
            "n_bootstraps": n_bootstraps,
            "seed": seed,
        }

        # Define analysis-relevant meta data
        self.attrs: dict[str, Any] = {"model": "BaseBootstrapper"}
        self.attrs.update(self._params)
        self.attrs.update(
            {
                "software": "xeofs",
                "version": __version__,
                "date": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            }
        )

        # Initialize the DataContainer to store the results
        self.data = DataContainer()

    @abstractmethod
    def fit(self, model):
        """Bootstrap a given model."""

        # VARIABLES ARE DEFINED IN THE SUBCLASS ACCORDING TO THE MODEL
        raise NotImplementedError


class EOFBootstrapper(_BaseBootstrapper, EOF):
    """Bootstrap a model to obtain significant modes and confidence intervals."""

    def __init__(self, n_bootstraps=20, seed=None):
        # Call the constructor of _BaseBootstrapper
        super().__init__(n_bootstraps=n_bootstraps, seed=seed)
        self.attrs.update({"model": "Bootstrapped EOF analysis"})

    def fit(self, model: EOF):
        """Bootstrap a given model."""

        self.model = model
        self.preprocessor = model.preprocessor
        sample_name = model.sample_name

        input_data = model.data["input_data"]
        n_samples = input_data.sample.size

        model_params = model.get_params()
        n_modes: int = model_params["n_modes"]
        n_bootstraps: int = self._params["n_bootstraps"]

        # Set seed for reproducibility
        rng = np.random.default_rng(self._params["seed"])

        # Bootstrap the model
        bst_expvar = []  # type: ignore
        bst_total_variance = []  # type: ignore
        bst_components = []  # type: ignore
        bst_scores = []  # type: ignore
        for i in trange(n_bootstraps):
            # Sample with replacement
            idx_rnd = rng.choice(n_samples, n_samples, replace=True)
            bst_data = input_data.isel({sample_name: idx_rnd})
            # We need to assign the sample coordinates of the real data
            # otherwise the transform() method will raise an error as it
            # tries to align the sample coordinates
            # with the coordinates of the bootstrapped (permutated) data
            bst_data = bst_data.assign_coords({sample_name: input_data[sample_name]})
            # Perform EOF analysis with the subsampled data
            # No scaling because we use the pre-scaled data from the model
            bst_model = EOF(n_modes=n_modes, standardize=False, use_coslat=False)
            bst_model.fit(bst_data, dim="sample")
            # Save results
            expvar = bst_model.data["explained_variance"]
            totvar = bst_model.data["total_variance"]
            components = bst_model.data["components"]
            scores = bst_model.transform(input_data, normalized=False)
            bst_expvar.append(expvar)
            bst_total_variance.append(totvar)
            bst_components.append(components)
            bst_scores.append(scores)

        # Concatenate the bootstrap results along a new dimension
        bst_expvar: DataArray = xr.concat(bst_expvar, dim="n")
        bst_total_variance: DataArray = xr.concat(bst_total_variance, dim="n")
        bst_components: DataArray = xr.concat(bst_components, dim="n")
        bst_scores: DataArray = xr.concat(bst_scores, dim="n")

        # Assign the bootstrap dimension coordinates
        coords_n = np.arange(1, n_bootstraps + 1)
        bst_expvar = bst_expvar.assign_coords(n=coords_n)
        bst_total_variance = bst_total_variance.assign_coords(n=coords_n)
        bst_components = bst_components.assign_coords(n=coords_n)
        bst_scores = bst_scores.assign_coords(n=coords_n)

        # Fix sign of individual components determined by correlation coefficients
        # for a given mode with all the individual bootstrap members
        # NOTE: we use scores as they have typically a lower dimensionality than components
        model_scores = model.data["scores"]
        corr = (
            (bst_scores * model_scores).mean("sample")
            / bst_scores.std("sample")
            / model_scores.std("sample")
        )
        signs = np.sign(corr)
        bst_components = bst_components * signs
        bst_scores = bst_scores * signs

        self.data.add(
            name="input_data", data=model.data["input_data"], allow_compute=False
        )
        self.data.add(name="components", data=bst_components)
        self.data.add(name="scores", data=bst_scores)
        self.data.add(name="norms", data=model.data["norms"])
        self.data.add(name="explained_variance", data=bst_expvar)
        self.data.add(name="total_variance", data=bst_total_variance)

        # Assign the same attributes as the original model
        self.data.set_attrs(self.attrs)



import importlib
import warnings
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Literal

import dask.base
import xarray as xr
from typing_extensions import Self

from ._version import __version__
from .utils.data_types import DataArray
from .utils.io import insert_placeholders, open_model_tree, write_model_tree
from .utils.xarray_utils import data_is_dask

# Ignore warnings from numpy casting with additional coordinates
warnings.filterwarnings("ignore", message=r"^invalid value encountered in cast*")

xr.set_options(keep_attrs=True)


class BaseModel(ABC):
    """
    Abstract base class for an xeofs model.

    Provides basic functionality for lazy model evaluation, serialization, deserialization and saving/loading models.

    """

    extra_modules = []
    uses_complex = False

    def __init__(self):
        # Define model parameters
        self._params = {}

        # Define analysis-relevant meta data
        self.attrs = {"model": "BaseModel"}
        self.attrs.update(
            {
                "software": "xeofs",
                "version": __version__,
                "date": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            }
        )
        self.attrs.update(self._params)

        # Ensure necessary non-core dependencies are available
        for module in self.extra_modules:
            self.check_needed_module(module)

    @abstractmethod
    def get_serialization_attrs(self) -> dict:
        """Get the attributes to serialize."""
        raise NotImplementedError

    def compute(self, **kwargs):
        """Compute and load delayed model results.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments to pass to `dask.compute()`.
        """
        # find and compute all dask arrays simultaneously to allow dask to optimize the
        # shared graph and avoid duplicate i/o and computations
        dt = self.serialize()

        data_objs = {
            k: v
            for k, v in dt.to_dict().items()
            if data_is_dask(v) and v.attrs.get("allow_compute", True)
        }

        (data_objs,) = dask.base.compute(data_objs, **kwargs)

        for k, v in data_objs.items():
            dt[k] = xr.DataTree(v)

        # then rebuild the trained model from the computed results
        self._deserialize_attrs(dt)

        self._post_compute()

    def _post_compute(self):
        pass

    def get_params(self) -> dict[str, Any]:
        """Get the model parameters."""
        return self._params

    def serialize(self) -> xr.DataTree:
        """Serialize a complete model with its preprocessor."""
        # Create a root node for this object with its params as attrs
        ds_root = xr.Dataset(attrs=dict(params=self.get_params()))
        dt = xr.DataTree(ds_root, name=type(self).__name__)

        # Retrieve the tree representation of each attached object, or set basic attrs
        for key, attr in self.get_serialization_attrs().items():
            if hasattr(attr, "serialize"):
                dt[key] = attr.serialize()
                dt.attrs[key] = "_is_tree"
            else:
                dt.attrs[key] = attr

        return dt

    def save(
        self,
        path: str,
        overwrite: bool = False,
        save_data: bool = False,
        engine: Literal["zarr", "netcdf4", "h5netcdf"] = "zarr",
        **kwargs,
    ):
        """Save the model.

        Parameters
        ----------
        path : str
            Path to save the model.
        overwrite: bool, default=False
            Whether or not to overwrite the existing path if it already exists.
            Ignored unless `engine="zarr"`.
        save_data : str
            Whether or not to save the full input data along with the fitted components.
        engine : {"zarr", "netcdf4", "h5netcdf"}, default="zarr"
            Xarray backend engine to use for writing the saved model.
        **kwargs
            Additional keyword arguments to pass to `DataTree.to_netcdf()` or `DataTree.to_zarr()`.

        """
        self.compute()

        dt = self.serialize()

        # Remove any raw data arrays at this stage
        if not save_data:
            dt = insert_placeholders(dt)

        if self.uses_complex and engine == "h5netcdf":
            kwargs = {"invalid_netcdf": True} | kwargs

        write_model_tree(dt, path, overwrite=overwrite, engine=engine, **kwargs)

    @classmethod
    def deserialize(cls, dt: xr.DataTree) -> Self:
        """Deserialize the model and its preprocessors from a DataTree."""
        # Recreate the model with parameters set by root level attrs
        model = cls(**dt.attrs["params"])
        model._deserialize_attrs(dt)
        return model

    def _deserialize_attrs(self, dt: xr.DataTree):
        """Set the necessary attributes of the model from a DataTree."""
        for key, attr in dt.attrs.items():
            if key == "params":
                continue
            elif attr == "_is_tree":
                deserialized_obj = getattr(self, str(key)).deserialize(dt[str(key)])
            else:
                deserialized_obj = attr
            setattr(self, str(key), deserialized_obj)

    @classmethod
    def load(
        cls,
        path: str,
        engine: Literal["zarr", "netcdf4", "h5netcdf"] = "zarr",
        **kwargs,
    ) -> Self:
        """Load a saved model.

        Parameters
        ----------
        path : str
            Path to the saved model.
        engine : {"zarr", "netcdf4", "h5netcdf"}, default="zarr"
            Xarray backend engine to use for reading the saved model.
        **kwargs
            Additional keyword arguments to pass to `open_datatree()`.

        Returns
        -------
        model : BaseModel
            The loaded model.

        """
        dt = open_model_tree(path, engine=engine, **kwargs)
        model = cls.deserialize(dt)
        return model

    def _validate_loaded_data(self, X: DataArray):
        """Optionally check the loaded data for placeholders."""
        pass

    def check_needed_module(self, module: str):
        """Check if a necessary non-core dependency is available."""
        try:
            importlib.import_module(module)
        except ImportError:
            raise ImportError(
                f"Additional module {module} is required for {self.__class__.__name__}."
            )



from typing import Any

import xarray as xr

from xeofs.utils.data_types import Dims


def assert_single_dataarray(da, name="object"):
    """Check if the given object is a DataArray.

    Args:
        da (DataArray): The object to check.
        name (str): The name of the object.

    Raises:
        TypeError: If the object is not a DataArray.
    """
    if not isinstance(da, xr.DataArray):
        raise TypeError(f"{name} must be a DataArray")


def assert_list_dataarrays(da_list, name="object"):
    """Check if the given object is a list of DataArrays.

    Args:
        da_list (list): The object to check.
        name (str): The name of the object.

    Raises:
        TypeError: If the object is not a list of DataArrays.
    """
    if not isinstance(da_list, list):
        raise TypeError(f"{name} must be a list of DataArrays")
    for da in da_list:
        assert_single_dataarray(da, name)


def assert_single_dataset(ds, name="object"):
    """Check if the given object is a Dataset.

    Args:
        ds (Dataset): The object to check.
        name (str): The name of the object.

    Raises:
        TypeError: If the object is not a Dataset.
    """
    if not isinstance(ds, xr.Dataset):
        raise TypeError(f"{name} must be a Dataset")


def assert_dataarray_or_dataset(da, name="object"):
    """Check if the given object is a DataArray or Dataset.

    Args:
        da (DataArray|Dataset): The object to check.
        name (str): The name of the object.

    Raises:
        TypeError: If the object is not a DataArray or Dataset.
    """
    if not isinstance(da, (xr.DataArray, xr.Dataset)):
        raise TypeError(f"{name} must be either a DataArray or Dataset")


def convert_to_dim_type(arg: Any) -> Dims:
    # Check for invalid types
    if not isinstance(arg, (str, tuple, list)):
        raise TypeError(f"Invalid input type: {type(arg).__name__}")

    # Check for invalid sequence elements
    if isinstance(arg, (tuple, list)) and not all(
        isinstance(item, str) for item in arg
    ):
        raise TypeError("Invalid sequence element type. All elements should be strings")

    if isinstance(arg, tuple):
        return arg
    elif isinstance(arg, list):
        return tuple(arg)
    else:
        return (arg,)


def validate_input_type(X) -> None:
    err_msg = "Invalid input type: {:}. Expected one of the following: DataArray, Dataset or list of these.".format(
        type(X).__name__
    )
    if isinstance(X, (xr.DataArray, xr.Dataset)):
        pass
    elif isinstance(X, (list, tuple)):
        if not all(isinstance(x, (xr.DataArray, xr.Dataset)) for x in X):
            raise TypeError(err_msg)
    else:
        raise TypeError(err_msg)


def assert_not_complex(da: xr.DataArray) -> None:
    if da.dtype == "complex":
        raise TypeError(
            "Invalid input type. This method does not support complex data types."
        )


def sanity_check_n_modes(n_modes: int | float | str) -> None:
    """Check if the number of modes is valid."""

    match n_modes:
        case int():
            if n_modes < 1:
                raise ValueError("If integer, n_modes must be greater than 0")
        case float():
            if not (0 < n_modes <= 1.0):
                raise ValueError("If float, n_modes must be in the range (0, 1]")
        case str():
            if n_modes not in ["all"]:
                raise ValueError("If string, n_modes must be 'all'")
        case _:
            raise TypeError("n_modes must be an integer, float or string.")



"""
EOF analysis (T-mode)
========================

EOF analysis in T-mode maximises the spatial variance.

Load packages and data:
"""

import matplotlib.pyplot as plt
import xarray as xr
from cartopy.crs import EqualEarth, PlateCarree
from matplotlib.gridspec import GridSpec

import xeofs as xe

sst = xr.tutorial.open_dataset("ersstv5")["sst"]

# %%
# Perform the actual analysis

model = xe.single.EOF(n_modes=5)
model.fit(sst, dim=("lat", "lon"))
expvar = model.explained_variance_ratio()
components = model.components()
scores = model.scores()

# %%
# Create figure showing the first two modes

proj = EqualEarth(central_longitude=180)
kwargs = {"cmap": "RdBu", "transform": PlateCarree()}

fig = plt.figure(figsize=(10, 8))
gs = GridSpec(3, 2, width_ratios=[2, 1])
ax0 = [fig.add_subplot(gs[i, 0], projection=proj) for i in range(3)]
ax1 = [fig.add_subplot(gs[i, 1]) for i in range(3)]

for i, (a0, a1) in enumerate(zip(ax0, ax1)):
    scores.sel(mode=i + 1).plot(ax=a0, **kwargs)
    a0.coastlines(color=".5")
    components.sel(mode=i + 1).plot(ax=a1)

    a0.set_xlabel("")

plt.tight_layout()
plt.savefig("eof-tmode.jpg")



class DimensionRenamer:
    """Rename dimensions of an xarray DataArray.

    Parameters
    ----------
    dim : str
        Name of the dimension to be renamed. Can be a dimension containing a MultiIndex.
    suffix : str
        Suffix to be added to the dimension name.

    """

    def __init__(self, dim, suffix):
        self.suffix = suffix
        self.dim = dim

    def fit(self, da):
        self.dims_mapping = {
            dim: dim + self.suffix for dim in da.coords[self.dim].coords.keys()
        }

    def transform(self, da):
        for old, new in self.dims_mapping.items():
            da = da.rename({old: new})
        return da

    def fit_transform(self, da):
        self.fit(da)
        return self.transform(da)

    def inverse_transform(self, da):
        for old, new in self.dims_mapping.items():
            da = da.rename({new: old})
        return da



import numpy as np
import pytest
import xarray as xr

from xeofs.cross import MCA

from ...utilities import data_is_dask, skip_if_missing_engine


@pytest.fixture
def mca_model():
    return MCA()


def test_initialization():
    mca = MCA()
    assert mca is not None


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    assert hasattr(mca_model, "preprocessor1")
    assert hasattr(mca_model, "preprocessor2")
    assert hasattr(mca_model, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit_empty_data(mca_model, dim):
    with pytest.raises(ValueError):
        mca_model.fit(xr.DataArray(), xr.DataArray(), dim)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit_invalid_dims(mca_model, mock_data_array, dim):
    with pytest.raises(ValueError):
        mca_model.fit(
            mock_data_array, mock_data_array, dim=("invalid_dim1", "invalid_dim2")
        )


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit_with_dataset(mca_model, mock_dataset, dim):
    mca_model.fit(mock_dataset, mock_dataset, dim)
    assert hasattr(mca_model, "preprocessor1")
    assert hasattr(mca_model, "preprocessor2")
    assert hasattr(mca_model, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit_with_dataarray_list(mca_model, mock_data_array_list, dim):
    mca_model.fit(mock_data_array_list, mock_data_array_list, dim)
    assert hasattr(mca_model, "preprocessor1")
    assert hasattr(mca_model, "preprocessor2")
    assert hasattr(mca_model, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_transform(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    result = mca_model.transform(X=mock_data_array, Y=mock_data_array)
    assert isinstance(result, list)
    assert isinstance(result[0], xr.DataArray)


@pytest.mark.parametrize("dim", [(("time",))])
def test_transform_unseen_data(mca_model, mock_data_array, dim):
    data = mock_data_array.isel(time=slice(0, 20))
    data_unseen = mock_data_array.isel(time=slice(21, None))

    mca_model.fit(data, data, dim)
    result = mca_model.transform(X=data_unseen, Y=data_unseen)
    assert isinstance(result, list)
    assert isinstance(result[0], xr.DataArray)
    # Check that unseen data can be transformed
    assert result[0].notnull().all()
    assert result[1].notnull().all()


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_inverse_transform(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    # Assuming mode as 1 for simplicity
    scores1 = mca_model.data["scores1"].isel(mode=1)
    scores2 = mca_model.data["scores2"].isel(mode=1)
    Xrec1, Xrec2 = mca_model.inverse_transform(scores1, scores2)
    assert isinstance(Xrec1, xr.DataArray)
    assert isinstance(Xrec2, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_squared_covariance_fraction(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    scf = mca_model.squared_covariance_fraction()
    assert isinstance(scf, xr.DataArray)
    assert all(scf <= 1), "Squared covariance fraction is greater than 1"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_covariance_fraction(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    cf = mca_model.covariance_fraction_CD95()
    assert isinstance(cf, xr.DataArray)
    assert all(cf <= 1), "Squared covariance fraction is greater than 1"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    components1, components2 = mca_model.components()
    feature_dims = tuple(set(mock_data_array.dims) - set(dim))
    assert isinstance(components1, xr.DataArray)
    assert isinstance(components2, xr.DataArray)
    assert set(components1.dims) == set(
        ("mode",) + feature_dims
    ), "Components1 does not have the right feature dimensions"
    assert set(components2.dims) == set(
        ("mode",) + feature_dims
    ), "Components2 does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_dataset(mca_model, mock_dataset, dim):
    mca_model.fit(mock_dataset, mock_dataset, dim)
    components1, components2 = mca_model.components()
    feature_dims = tuple(set(mock_dataset.dims) - set(dim))
    assert isinstance(components1, xr.Dataset)
    assert isinstance(components2, xr.Dataset)
    assert set(components1.data_vars) == set(
        mock_dataset.data_vars
    ), "Components does not have the same data variables as the input Dataset"
    assert set(components2.data_vars) == set(
        mock_dataset.data_vars
    ), "Components does not have the same data variables as the input Dataset"
    assert set(components1.dims) == set(
        ("mode",) + feature_dims
    ), "Components does not have the right feature dimensions"
    assert set(components2.dims) == set(
        ("mode",) + feature_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_dataarray_list(mca_model, mock_data_array_list, dim):
    mca_model.fit(mock_data_array_list, mock_data_array_list, dim)
    components1, components2 = mca_model.components()
    feature_dims = [tuple(set(data.dims) - set(dim)) for data in mock_data_array_list]
    assert isinstance(components1, list)
    assert isinstance(components2, list)
    assert len(components1) == len(mock_data_array_list)
    assert len(components2) == len(mock_data_array_list)
    assert isinstance(components1[0], xr.DataArray)
    assert isinstance(components2[0], xr.DataArray)
    for comp, feat_dims in zip(components1, feature_dims):
        assert set(comp.dims) == set(
            ("mode",) + feat_dims
        ), "Components1 does not have the right feature dimensions"
    for comp, feat_dims in zip(components2, feature_dims):
        assert set(comp.dims) == set(
            ("mode",) + feat_dims
        ), "Components2 does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    scores1, scores2 = mca_model.scores()
    assert isinstance(scores1, xr.DataArray)
    assert isinstance(scores2, xr.DataArray)
    assert set(scores1.dims) == set(
        (dim + ("mode",))
    ), "Scores1 does not have the right dimensions"
    assert set(scores2.dims) == set(
        (dim + ("mode",))
    ), "Scores2 does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_dataset(mca_model, mock_dataset, dim):
    mca_model.fit(mock_dataset, mock_dataset, dim)
    scores1, scores2 = mca_model.scores()
    assert isinstance(scores1, xr.DataArray)
    assert isinstance(scores2, xr.DataArray)
    assert set(scores1.dims) == set(
        (dim + ("mode",))
    ), "Scores1 does not have the right dimensions"
    assert set(scores2.dims) == set(
        (dim + ("mode",))
    ), "Scores2 does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_dataarray_list(mca_model, mock_data_array_list, dim):
    mca_model.fit(mock_data_array_list, mock_data_array_list, dim)
    scores1, scores2 = mca_model.scores()
    assert isinstance(scores1, xr.DataArray)
    assert isinstance(scores2, xr.DataArray)
    assert set(scores1.dims) == set(
        (dim + ("mode",))
    ), "Scores1 does not have the right dimensions"
    assert set(scores2.dims) == set(
        (dim + ("mode",))
    ), "Scores2 does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_homogeneous_patterns(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    patterns, pvals = mca_model.homogeneous_patterns()
    assert isinstance(patterns[0], xr.DataArray)
    assert isinstance(patterns[1], xr.DataArray)
    assert isinstance(pvals[0], xr.DataArray)
    assert isinstance(pvals[1], xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_heterogeneous_patterns(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    patterns, pvals = mca_model.heterogeneous_patterns()
    assert isinstance(patterns[0], xr.DataArray)
    assert isinstance(patterns[1], xr.DataArray)
    assert isinstance(pvals[0], xr.DataArray)
    assert isinstance(pvals[1], xr.DataArray)


@pytest.mark.parametrize(
    "dim, compute",
    [
        (("time",), True),
        (("lat", "lon"), True),
        (("lon", "lat"), True),
        (("time",), False),
        (("lat", "lon"), False),
        (("lon", "lat"), False),
    ],
)
def test_compute(mock_dask_data_array, dim, compute):
    mca_model = MCA(n_modes=10, compute=compute, n_pca_modes=10)
    mca_model.fit(mock_dask_data_array, mock_dask_data_array, (dim))

    if compute:
        assert not data_is_dask(mca_model.data["singular_values"])
        assert not data_is_dask(mca_model.data["components1"])
        assert not data_is_dask(mca_model.data["components2"])

    else:
        assert data_is_dask(mca_model.data["singular_values"])
        assert data_is_dask(mca_model.data["components1"])
        assert data_is_dask(mca_model.data["components2"])


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
@pytest.mark.parametrize("engine", ["h5netcdf", "netcdf4", "zarr"])
def test_save_load(dim, mock_data_array, tmp_path, engine):
    """Test save/load methods in MCA class, ensuring that we can
    roundtrip the model and get the same results when transforming
    data."""
    skip_if_missing_engine(engine)

    original = MCA()
    original.fit(mock_data_array, mock_data_array, dim)

    # Save the EOF model
    original.save(tmp_path / "mca", engine=engine)

    # Check that the EOF model has been saved
    assert (tmp_path / "mca").exists()

    # Recreate the model from saved file
    loaded = MCA.load(tmp_path / "mca", engine=engine)

    # Check that the params and DataContainer objects match
    assert original.get_params() == loaded.get_params()
    assert all([key in loaded.data for key in original.data])
    for key in original.data:
        if original.data._allow_compute[key]:
            assert loaded.data[key].equals(original.data[key])
        else:
            # but ensure that input data is not saved by default
            assert loaded.data[key].size <= 1
            assert loaded.data[key].attrs["placeholder"] is True

    # Test that the recreated model can be used to transform new data
    assert np.allclose(
        original.transform(mock_data_array, mock_data_array),
        loaded.transform(mock_data_array, mock_data_array),
    )

    # The loaded model should also be able to inverse_transform new data
    assert np.allclose(
        original.inverse_transform(*original.scores()),
        loaded.inverse_transform(*loaded.scores()),
    )


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_serialize_deserialize_dataarray(dim, mock_data_array):
    """Test roundtrip serialization when the model is fit on a DataArray."""
    model = MCA()
    model.fit(mock_data_array, mock_data_array, dim)
    dt = model.serialize()
    rebuilt_model = MCA.deserialize(dt)
    assert np.allclose(
        model.transform(mock_data_array), rebuilt_model.transform(mock_data_array)
    )


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_serialize_deserialize_dataset(dim, mock_dataset):
    """Test roundtrip serialization when the model is fit on a Dataset."""
    model = MCA()
    model.fit(mock_dataset, mock_dataset, dim)
    dt = model.serialize()
    rebuilt_model = MCA.deserialize(dt)
    assert np.allclose(
        model.transform(mock_dataset), rebuilt_model.transform(mock_dataset)
    )



import numpy as np
import numba


@numba.njit(fastmath=True)
def kernel_weights_nb(distance, bandwidth, kernel):
    if kernel == "bisquare":
        return _bisquare_nb(distance, bandwidth)
    elif kernel == "gaussian":
        return _gaussian_nb(distance, bandwidth)
    elif kernel == "exponential":
        return _exponential_nb(distance, bandwidth)
    else:
        raise ValueError(
            f"Invalid kernel: {kernel}. Must be one of ['bisquare', 'gaussian', 'exponential']."
        )


@numba.njit(fastmath=True)
def _bisquare_nb(distance, bandwidth):
    weights = (1 - (distance / bandwidth) ** 2) ** 2
    return np.where(distance <= bandwidth, weights, 0)


@numba.njit(fastmath=True)
def _gaussian_nb(distance, bandwidth):
    return np.exp(-0.5 * (distance / bandwidth) ** 2)


@numba.njit(fastmath=True)
def _exponential_nb(distance, bandwidth):
    return np.exp(-0.5 * (distance / bandwidth))



import numpy as np
import xarray as xr
from typing_extensions import Self

from ..utils.data_types import (
    DataArray,
    Dims,
    DimsList,
)
from .transformer import Transformer


class Concatenator(Transformer):
    """Concatenate a list of DataArrays along the feature dimensions."""

    def __init__(self, sample_name: str = "sample", feature_name: str = "feature"):
        super().__init__(sample_name, feature_name)

        self.n_data = None
        self.n_features = []
        self.coords_in = {}

    def get_serialization_attrs(self) -> dict:
        return dict(
            n_data=self.n_data,
            n_features=self.n_features,
            coords_in=self.coords_in,
        )

    def fit(
        self,
        X: list[DataArray],
        sample_dims: Dims | None = None,
        feature_dims: DimsList | None = None,
    ) -> Self:
        # Check that all inputs are DataArrays
        if not all([isinstance(data, DataArray) for data in X]):
            raise ValueError("Input must be a list of DataArrays")

        # Check that all inputs have shape 2
        if not all([len(data.dims) == 2 for data in X]):
            raise ValueError("Input DataArrays must have shape 2")

        # Check that all inputs have the same sample_name and feature_name
        if not all([data.dims == (self.sample_name, self.feature_name) for data in X]):
            raise ValueError("Input DataArrays must have the same dimensions")

        self.n_data = len(X)

        # Set input feature coordinates, using dict for easier serialization
        self.coords_in = {
            str(i): data.coords[self.feature_name] for i, data in enumerate(X)
        }
        self.n_features = [coord.size for coord in self.coords_in.values()]

        return self

    def transform(self, X: list[DataArray]) -> DataArray:
        # Test whether the input list has same length as the number of stackers
        if len(X) != self.n_data:
            raise ValueError(
                f"Invalid input. Number of DataArrays ({len(X)}) does not match the number of fitted DataArrays ({self.n_data})."
            )

        reindexed_data_list: list[DataArray] = []

        idx_range = np.cumsum([0] + self.n_features)
        for i, data in enumerate(X):
            # Create dummy feature coordinates for DataArray
            new_coords = np.arange(idx_range[i], idx_range[i + 1])

            # Replace original feature coordinates with dummy coordinates
            data = data.drop_vars(self.feature_name)
            reindexed = data.assign_coords({self.feature_name: new_coords})

            reindexed_data_list.append(reindexed)

        X_concat: DataArray = xr.concat(reindexed_data_list, dim=self.feature_name)
        self.coords_out = X_concat.coords[self.feature_name]

        return X_concat

    def fit_transform(
        self,
        X: list[DataArray],
        sample_dims: Dims | None = None,
        feature_dims: DimsList | None = None,
    ) -> DataArray:
        return self.fit(X, sample_dims, feature_dims).transform(X)

    def _split_dataarray_into_list(self, data: DataArray) -> list[DataArray]:
        feature_name = self.feature_name
        data_list: list[DataArray] = []

        idx_range = np.cumsum([0] + self.n_features)
        for i, coords in enumerate(self.coords_in.values()):
            # Create dummy feature coordinates for DataArray
            features = np.arange(idx_range[i], idx_range[i + 1])
            # Select the features corresponding to the current DataArray
            sub_selection = data.sel({feature_name: features})
            # Replace dummy feature coordinates with original feature coordinates
            sub_selection = sub_selection.assign_coords({feature_name: coords})
            data_list.append(sub_selection)

        return data_list

    def inverse_transform_data(self, X: DataArray) -> list[DataArray]:
        """Reshape the 2D data (sample x feature) back into its original shape."""
        return self._split_dataarray_into_list(X)

    def inverse_transform_components(self, X: DataArray) -> list[DataArray]:
        """Reshape the 2D components (sample x feature) back into its original shape."""
        return self._split_dataarray_into_list(X)

    def inverse_transform_scores(self, X: DataArray) -> DataArray:
        """Reshape the 2D scores (sample x mode) back into its original shape."""
        return X

    def inverse_transform_scores_unseen(self, X: DataArray) -> DataArray:
        """Reshape the 2D scores (sample x mode) back into its original shape."""
        return X
